@@ -7,21 +7,10 @@
 
 #include "scxtest/scx_test.h"
 #include <scx/common.bpf.h>
+#include <lib/arena_map.h>
 #include <lib/sdt_task.h>
 #include <scx/arena_userspace_interrop.bpf.h>
 
-struct {
-	__uint(type, BPF_MAP_TYPE_ARENA);
-	__uint(map_flags, BPF_F_MMAPABLE);
-#if defined(__TARGET_ARCH_arm64) || defined(__aarch64__)
-	__uint(max_entries, 1 << 16); /* number of pages */
-        __ulong(map_extra, (1ull << 32)); /* start of mmap() region */
-#else
-	__uint(max_entries, 1 << 20); /* number of pages */
-        __ulong(map_extra, (1ull << 44)); /* start of mmap() region */
-#endif
-} arena __weak SEC(".maps");
-
 struct scx_alloc_stack __arena *prealloc_stack;
 
 /*
@@ -1010,7 +999,7 @@ int header_set_order(scx_buddy_chunk_t *chunk, u64 offset, u8 order)
 
 	if (offset >= SCX_BUDDY_CHUNK_ITEMS) {
 		bpf_printk("setting order of invalid offset");
-		return EINVAL;
+		return -EINVAL;
 	}
 
 	if (offset & 0x1)
