@@ -13,7 +13,7 @@
  * to be dispatched in the proper order.
  *
  * Messages between the BPF component and the user-space scheduler are passed
- * using BPF_MAP_TYPE_RINGBUFFER / BPF_MAP_TYPE_USER_RINGBUF maps: @queued for
+ * using BPF_MAP_TYPE_RINGBUF / BPF_MAP_TYPE_USER_RINGBUF maps: @queued for
  * the messages sent by the BPF dispatcher to the user-space scheduler and
  * @dispatched for the messages sent by the user-space scheduler to the BPF
  * dispatcher.
@@ -34,6 +34,7 @@
 #include <scx/common.bpf.h>
 #endif
 
+#include <scx/percpu.bpf.h>
 #include "intf.h"
 
 char _license[] SEC("license") = "GPL";
@@ -68,9 +69,9 @@ u64 usersched_last_run_at; /* Timestamp of the last user-space scheduler executi
 static u64 nr_cpu_ids; /* Maximum possible CPU number */
 
 /*
- * Switch all tasks or SCHED_EXT tasks.
+ * Default task time slice.
  */
-const volatile bool switch_partial;
+const volatile u64 slice_ns;
 
 /*
  * Number of tasks that are queued for scheduling.
@@ -101,7 +102,7 @@ volatile u64 nr_user_dispatches, nr_kernel_dispatches,
 /* Failure statistics */
 volatile u64 nr_failed_dispatches, nr_sched_congested;
 
- /* Report additional debugging information */
+/* Report additional debugging information */
 const volatile bool debug;
 
 /* Rely on the in-kernel idle CPU selection policy */
@@ -118,24 +119,6 @@ const volatile bool builtin_idle;
  */
 const volatile bool smt_enabled = true;
 
-/*
- * Allocate/re-allocate a new cpumask.
- */
-static int calloc_cpumask(struct bpf_cpumask **p_cpumask)
-{
-	struct bpf_cpumask *cpumask;
-
-	cpumask = bpf_cpumask_create();
-	if (!cpumask)
-		return -ENOMEM;
-
-	cpumask = bpf_kptr_xchg(p_cpumask, cpumask);
-	if (cpumask)
-		bpf_cpumask_release(cpumask);
-
-	return 0;
-}
-
 /*
  * Maximum amount of tasks queued between kernel and user-space at a certain
  * time.
@@ -153,7 +136,7 @@ static int calloc_cpumask(struct bpf_cpumask **p_cpumask)
 /*
  * The map containing tasks that are queued to user space from the kernel.
  *
- * This map is drained by the user space scheduler.
+ * This map is drained by the user-space scheduler.
  */
 struct {
 	__uint(type, BPF_MAP_TYPE_RINGBUF);
@@ -173,42 +156,12 @@ struct {
 				sizeof(struct dispatched_task_ctx));
 } dispatched SEC(".maps");
 
-/*
- * Per-CPU context.
- */
-struct cpu_ctx {
-	struct bpf_cpumask __kptr *l2_cpumask;
-	struct bpf_cpumask __kptr *l3_cpumask;
-};
-
-struct {
-	__uint(type, BPF_MAP_TYPE_PERCPU_ARRAY);
-	__type(key, u32);
-	__type(value, struct cpu_ctx);
-	__uint(max_entries, 1);
-} cpu_ctx_stor SEC(".maps");
-
-/*
- * Return a CPU context.
- */
-struct cpu_ctx *try_lookup_cpu_ctx(s32 cpu)
-{
-	const u32 idx = 0;
-	return bpf_map_lookup_percpu_elem(&cpu_ctx_stor, &idx, cpu);
-}
-
 /*
  * Per-task local storage.
  *
  * This contain all the per-task information used internally by the BPF code.
  */
 struct task_ctx {
-	/*
-	 * Temporary cpumask for calculating scheduling domains.
-	 */
-	struct bpf_cpumask __kptr *l2_cpumask;
-	struct bpf_cpumask __kptr *l3_cpumask;
-
 	/*
 	 * Timestamp since last time the task ran on a CPU.
 	 */
@@ -223,6 +176,11 @@ struct task_ctx {
 	 * Execution time (in nanoseconds) since the last sleep event.
 	 */
 	u64 exec_runtime;
+
+	/*
+	 * Task generation counter to detect duplicate enqueues.
+	 */
+	u64 enq_cnt;
 };
 
 /* Map that contains task-local storage. */
@@ -269,7 +227,7 @@ struct {
  * Time period of the scheduler heartbeat, used to periodically kick the
  * user-space scheduler and check if there is any pending activity.
  */
-#define USERSCHED_TIMER_NS (NSEC_PER_SEC / 10)
+#define USERSCHED_TIMER_NS	NSEC_PER_SEC
 
 /*
  * Return true if the target task @p is the user-space scheduler.
@@ -359,6 +317,9 @@ static bool test_and_clear_usersched_needed(void)
  */
 static bool usersched_has_pending_tasks(void)
 {
+	if (test_and_clear_usersched_needed())
+		return true;
+
 	if (nr_scheduled)
 		return true;
 
@@ -378,6 +339,55 @@ static u64 cpu_to_dsq(s32 cpu)
 	return (u64)cpu;
 }
 
+/*
+ * Return true if @this_cpu and @that_cpu are in the same LLC, false
+ * otherwise.
+ */
+static inline bool cpus_share_cache(s32 this_cpu, s32 that_cpu)
+{
+        if (this_cpu == that_cpu)
+                return true;
+
+	return cpu_llc_id(this_cpu) == cpu_llc_id(that_cpu);
+}
+
+/*
+ * Return true if @this_cpu is faster than @that_cpu, false otherwise.
+ */
+static inline bool is_cpu_faster(s32 this_cpu, s32 that_cpu)
+{
+        if (this_cpu == that_cpu)
+                return false;
+
+	return cpu_priority(this_cpu) > cpu_priority(that_cpu);
+}
+
+/*
+ * Return true if @cpu is a fully-idle SMT core, false otherwise.
+ */
+static inline bool is_smt_idle(s32 cpu)
+{
+	const struct cpumask *idle_smtmask;
+        bool is_idle;
+
+	if (!smt_enabled)
+		return true;
+
+	idle_smtmask = scx_bpf_get_idle_smtmask();
+        is_idle = bpf_cpumask_test_cpu(cpu, idle_smtmask);
+        scx_bpf_put_cpumask(idle_smtmask);
+
+	return is_idle;
+}
+
+/*
+ * Return true on a wake-up event, false otherwise.
+ */
+static inline bool is_wakeup(u64 wake_flags)
+{
+	return wake_flags & SCX_WAKE_TTWU;
+}
+
 /*
  * Find an idle CPU in the system for the task.
  *
@@ -387,168 +397,65 @@ static u64 cpu_to_dsq(s32 cpu)
  * to handle these mistakes in favor of a more efficient response and a reduced
  * scheduling overhead.
  */
-static s32 pick_idle_cpu(const struct task_struct *p, s32 prev_cpu)
+static s32 pick_idle_cpu(struct task_struct *p, s32 prev_cpu, u64 wake_flags)
 {
-	const struct cpumask *idle_smtmask;
-	struct bpf_cpumask *l2_domain, *l3_domain;
-	struct bpf_cpumask *l2_mask, *l3_mask;
-	struct task_ctx *tctx;
-	struct cpu_ctx *cctx;
-	s32 cpu;
+	s32 cpu, this_cpu = bpf_get_smp_processor_id();
+	bool is_this_cpu_allowed = bpf_cpumask_test_cpu(this_cpu, p->cpus_ptr);
 
 	/*
 	 * For tasks that can run only on a single CPU, we can simply verify if
 	 * their only allowed CPU is still idle.
 	 */
-	if (p->nr_cpus_allowed == 1 || is_migration_disabled(p)) {
+	if (p->nr_cpus_allowed == 1) {
 		if (scx_bpf_test_and_clear_cpu_idle(prev_cpu))
 			return prev_cpu;
 
 		return -EBUSY;
 	}
 
-	tctx = try_lookup_task_ctx(p);
-	if (!tctx)
-		return -ENOENT;
-
-	cctx = try_lookup_cpu_ctx(prev_cpu);
-	if (!cctx)
-		return -ENOENT;
-
 	/*
-	 * Acquire the CPU masks to determine the idle CPUs in the system.
+	 * On wakeup if the waker's CPU is faster than the wakee's CPU, try
+	 * to move the wakee closer to the waker.
+	 *
+	 * In presence of hybrid cores this helps to naturally migrate
+	 * tasks over to the faster cores.
 	 */
-	idle_smtmask = scx_bpf_get_idle_smtmask();
-
-	/*
-	 * Scheduling domains of the previously used CPU.
-	 */
-	l2_domain = cctx->l2_cpumask;
-	if (!l2_domain)
-		l2_domain = (struct bpf_cpumask *)p->cpus_ptr;
-
-	l3_domain = cctx->l3_cpumask;
-	if (!l3_domain)
-		l3_domain = (struct bpf_cpumask *)p->cpus_ptr;
-
-	if (p->nr_cpus_allowed == nr_cpu_ids) {
-		l2_mask = l2_domain;
-		l3_mask = l3_domain;
-	} else {
+	if (is_wakeup(wake_flags) &&
+	    is_cpu_faster(this_cpu, prev_cpu) && is_this_cpu_allowed) {
 		/*
-		 * Determine the cache domain as the intersection of the
-		 * task's primary cpumask and the cache domain mask of the
-		 * previously used CPU (ignore if the cache cpumask
-		 * completely overlaps with the task's cpumask).
+		 * If both the waker's CPU and the wakee's CPU are in the
+		 * same LLC and the wakee's CPU is a fully idle SMT core,
+		 * don't migrate.
 		 */
-		l2_mask = tctx->l2_cpumask;
-		if (!l2_mask) {
-			scx_bpf_error("L2 cpumask not initialized");
-			cpu = -ENOENT;
-			goto out_put_cpumask;
-		}
-		if (!bpf_cpumask_and(l2_mask, p->cpus_ptr, cast_mask(l2_domain)))
-			l2_mask = NULL;
+		if (cpus_share_cache(this_cpu, prev_cpu) &&
+		    is_smt_idle(prev_cpu) &&
+		    scx_bpf_test_and_clear_cpu_idle(prev_cpu))
+			return prev_cpu;
 
-		l3_mask = tctx->l3_cpumask;
-		if (!l3_mask) {
-			scx_bpf_error("L3 cpumask not initialized");
-			cpu = -ENOENT;
-			goto out_put_cpumask;
-		}
-		if (!bpf_cpumask_and(l3_mask, p->cpus_ptr, cast_mask(l3_domain)))
-			l3_mask = NULL;
+		prev_cpu = this_cpu;
 	}
 
 	/*
-	 * Find the best idle CPU, prioritizing full idle cores in SMT systems.
+	 * Fallback to the old API if the kernel doesn't support
+	 * scx_bpf_select_cpu_and().
+	 *
+	 * This is required to support kernels <= 6.16.
 	 */
-	if (smt_enabled) {
-		/*
-		 * If the task can still run on the previously used CPU and
-		 * it's a full-idle core, keep using it.
-		 */
-		if (bpf_cpumask_test_cpu(prev_cpu, idle_smtmask) &&
-		    scx_bpf_test_and_clear_cpu_idle(prev_cpu)) {
-			cpu = prev_cpu;
-			goto out_put_cpumask;
-		}
+	if (!bpf_ksym_exists(scx_bpf_select_cpu_and)) {
+		bool is_idle = false;
 
-		/*
-		 * Search for any full-idle CPU in the task domain that shares
-		 * the same L2 cache.
-		 */
-		if (l2_mask) {
-			cpu = scx_bpf_pick_idle_cpu(cast_mask(l2_mask), SCX_PICK_IDLE_CORE);
-			if (cpu >= 0)
-				goto out_put_cpumask;
-		}
+		if (!wake_flags)
+			return -EBUSY;
 
-		/*
-		 * Search for any full-idle CPU in the task domain that shares
-		 * the same L3 cache.
-		 */
-		if (l3_mask) {
-			cpu = scx_bpf_pick_idle_cpu(cast_mask(l3_mask), SCX_PICK_IDLE_CORE);
-			if (cpu >= 0)
-				goto out_put_cpumask;
-		}
+		cpu = scx_bpf_select_cpu_dfl(p, prev_cpu, wake_flags, &is_idle);
 
-		/*
-		 * Otherwise, search for another usable full-idle core.
-		 */
-		cpu = scx_bpf_pick_idle_cpu(p->cpus_ptr, SCX_PICK_IDLE_CORE);
-		if (cpu >= 0)
-			goto out_put_cpumask;
+		return is_idle ? cpu : -EBUSY;
 	}
 
 	/*
-	 * If a full-idle core can't be found (or if this is not an SMT system)
-	 * try to re-use the same CPU, even if it's not in a full-idle core.
+	 * Pick any idle CPU usable by the task.
 	 */
-	if (scx_bpf_test_and_clear_cpu_idle(prev_cpu)) {
-		cpu = prev_cpu;
-		goto out_put_cpumask;
-	}
-
-	/*
-	 * Search for any idle CPU in the primary domain that shares the same
-	 * L2 cache.
-	 */
-	if (l2_mask) {
-		cpu = scx_bpf_pick_idle_cpu(cast_mask(l2_mask), 0);
-		if (cpu >= 0)
-			goto out_put_cpumask;
-	}
-
-	/*
-	 * Search for any idle CPU in the primary domain that shares the same
-	 * L3 cache.
-	 */
-	if (l3_mask) {
-		cpu = scx_bpf_pick_idle_cpu(cast_mask(l3_mask), 0);
-		if (cpu >= 0)
-			goto out_put_cpumask;
-	}
-
-	/*
-	 * If all the previous attempts have failed, try to use any idle CPU in
-	 * the system.
-	 */
-	cpu = scx_bpf_pick_idle_cpu(p->cpus_ptr, 0);
-	if (cpu >= 0)
-		goto out_put_cpumask;
-
-	/*
-	 * If all the previous attempts have failed, dispatch the task to the
-	 * first CPU that will become available.
-	 */
-	cpu = -EBUSY;
-
-out_put_cpumask:
-	scx_bpf_put_cpumask(idle_smtmask);
-
-	return cpu;
+	return scx_bpf_select_cpu_and(p, prev_cpu, wake_flags, p->cpus_ptr, 0);
 }
 
 /*
@@ -580,8 +487,9 @@ static void kick_task_cpu(const struct task_struct *p, s32 cpu)
  */
 static void dispatch_task(const struct dispatched_task_ctx *task)
 {
+	struct task_ctx *tctx;
 	struct task_struct *p;
-	s32 prev_cpu;
+	s32 prev_cpu, cpu = task->cpu;
 
 	/* Ignore entry if the task doesn't exist anymore */
 	p = bpf_task_from_pid(task->pid);
@@ -597,135 +505,70 @@ static void dispatch_task(const struct dispatched_task_ctx *task)
 		scx_bpf_dsq_insert_vtime(p, SHARED_DSQ,
 					 task->slice_ns, task->vtime, task->flags);
 		kick_task_cpu(p, prev_cpu);
-
 		goto out_release;
 	}
 
 	/*
-	 * If the target CPU selected by the user-space scheduler is not
-	 * valid, dispatch it to the SHARED_DSQ, independently on what the
-	 * user-space scheduler has decided.
-	 */
-	if (!bpf_cpumask_test_cpu(task->cpu, p->cpus_ptr)) {
-		scx_bpf_dsq_insert_vtime(p, SHARED_DSQ,
-					 task->slice_ns, task->vtime, task->flags);
-		__sync_fetch_and_add(&nr_bounce_dispatches, 1);
-		kick_task_cpu(p, prev_cpu);
-
-		goto out_release;
-	}
-
-	/*
-	 * Dispatch a task to a target CPU selected by the user-space
-	 * scheduler.
-	 */
-	scx_bpf_dsq_insert_vtime(p, cpu_to_dsq(task->cpu),
-				 task->slice_ns, task->vtime, task->flags);
-	__sync_fetch_and_add(&nr_user_dispatches, 1);
-
-	/*
-	 * If the cpumask is not valid anymore, ignore the dispatch event.
+	 * Dispatch the task to the target CPU selected by the
+	 * user-space scheduler.
 	 *
-	 * This can happen if the task has changed its affinity and the
-	 * target CPU has become invalid. In this case cancelling the
-	 * dispatch allows to prevent potential stalls in the scheduler,
-	 * since the task will be re-enqueued by the core sched-ext code,
-	 * potentially selecting a different CPU.
+	 * However, if the target CPU is not valid (due to affinity
+	 * constraints), keep the task on the previously used CPU,
+	 * overriding the user-space scheduler decision.
 	 */
 	if (!bpf_cpumask_test_cpu(task->cpu, p->cpus_ptr)) {
+		cpu = prev_cpu;
+		__sync_fetch_and_add(&nr_bounce_dispatches, 1);
+	} else {
+		__sync_fetch_and_add(&nr_user_dispatches, 1);
+	}
+	scx_bpf_dsq_insert_vtime(p, cpu_to_dsq(cpu),
+				 task->slice_ns, task->vtime, task->flags);
+
+	/*
+	 * If the task was dequeued while still in the user-space
+	 * scheduler, this dispatch can be ignored.
+	 *
+	 * Another enqueue event for the same task will be received later.
+	 */
+	tctx = try_lookup_task_ctx(p);
+	if (!tctx || tctx->enq_cnt > task->enq_cnt) {
 		scx_bpf_dispatch_cancel();
 		__sync_fetch_and_add(&nr_cancel_dispatches, 1);
-
 		goto out_release;
 	}
 
-	scx_bpf_kick_cpu(task->cpu, SCX_KICK_IDLE);
+	/*
+	 * CPU selected by the user-space scheduler is valid, kick it.
+	 */
+	scx_bpf_kick_cpu(cpu, SCX_KICK_IDLE);
 
 out_release:
 	bpf_task_release(p);
 }
 
 /*
- * Return true if the waker commits to release the CPU after waking up @p,
- * false otherwise.
+ * Return true it's safe to dispatch directly on @cpu, false otherwise.
  */
-static bool is_wake_sync(u64 wake_flags)
+static bool can_direct_dispatch(s32 cpu)
 {
-	const struct task_struct *current = (void *)bpf_get_current_task_btf();
-
-	return (wake_flags & SCX_WAKE_SYNC) && !(current->flags & PF_EXITING);
-}
-
-/*
- * Try to dispatch a task directly on an idle CPU.
- *
- * On success, return the idle CPU where the task has been dispatched, or
- * return a negative value otherwise.
- */
-static s32 try_direct_dispatch(struct task_struct *p,
-			       s32 prev_cpu, u64 enq_flags, bool *dispatched)
-{
-	s32 cpu;
-
-	*dispatched = false;
-
-	/*
-	 * If built-in idle CPU policy is not enabled completely delegate
-	 * the idle selection policy to user-space and keep re-using the
-	 * same CPU here.
-	 */
-	if (!builtin_idle)
-		return -EBUSY;
-
-	/*
-	 * Pick the idle CPU closest to prev_cpu usable by the task.
-	 */
-	cpu = pick_idle_cpu(p, prev_cpu);
-	if (cpu < 0)
-		return cpu;
-
-	/*
-	 * If we picked an invalid CPU for the task, give up and ignore
-	 * direct dispatch.
-	 */
-	if (!bpf_cpumask_test_cpu(cpu, p->cpus_ptr)) {
-		/*
-		 * We selected a CPU that the task cannot run on, which
-		 * causes it to become idle-locked, so it can't be chosen
-		 * by other tasks and may remain unused for a long time.
-		 *
-		 * To avoid wasting CPU time, we kick the CPU to let it
-		 * pull a task from the SHARED_DSQ. If a task is available,
-		 * it runs for a slice, then returns idle again, refreshing
-		 * its state and becoming eligible again as an idle CPU. If
-		 * no task is available, it will continue running the idle
-		 * task, refreshing its idle state and rejoining the pool
-		 * of idle CPUs.
-		 */
-		scx_bpf_kick_cpu(cpu, SCX_KICK_IDLE);
-		return -EBUSY;
-	}
-
-	/*
-	 * Perform direct dispatch only if the SHAREQ_DSQ is empty,
-	 * otherwise we may risk to starve the tasks waiting in the
-	 * SHARED_DSQ.
-	 */
-	if (!scx_bpf_dsq_nr_queued(SHARED_DSQ) && !scx_bpf_dsq_nr_queued(cpu_to_dsq(cpu))) {
-		scx_bpf_dsq_insert_vtime(p, cpu_to_dsq(cpu),
-					 SCX_SLICE_DFL, p->scx.dsq_vtime, enq_flags);
-		__sync_fetch_and_add(&nr_kernel_dispatches, 1);
-		*dispatched = true;
-	}
-
-	return cpu;
+	return !scx_bpf_dsq_nr_queued(SHARED_DSQ) &&
+	       !scx_bpf_dsq_nr_queued(cpu_to_dsq(cpu));
 }
 
 s32 BPF_STRUCT_OPS(rustland_select_cpu, struct task_struct *p, s32 prev_cpu,
 		   u64 wake_flags)
 {
-	bool dispatched = false;
-	s32 cpu;
+	s32 cpu, this_cpu = bpf_get_smp_processor_id();
+	bool is_this_cpu_allowed = bpf_cpumask_test_cpu(this_cpu, p->cpus_ptr);
+
+	/*
+	 * Make sure @prev_cpu is usable, otherwise try to move close to
+	 * the waker's CPU. If the waker's CPU is also not usable, then
+	 * pick the first usable CPU.
+	 */
+	if (!bpf_cpumask_test_cpu(prev_cpu, p->cpus_ptr))
+		prev_cpu = is_this_cpu_allowed ? this_cpu : bpf_cpumask_first(p->cpus_ptr);
 
 	/*
 	 * Scheduler is dispatched directly in .dispatch() when needed, so
@@ -734,16 +577,32 @@ s32 BPF_STRUCT_OPS(rustland_select_cpu, struct task_struct *p, s32 prev_cpu,
 	if (is_usersched_task(p))
 		return prev_cpu;
 
-	cpu = try_direct_dispatch(p, prev_cpu, 0, &dispatched);
-	if (cpu >= 0)
+	/*
+	 * If built-in idle CPU policy is not enabled, completely delegate
+	 * the idle selection policy to user-space and keep re-using the
+	 * same CPU here.
+	 */
+	if (!builtin_idle)
+		return prev_cpu;
+
+	/*
+	 * Pick the idle CPU closest to @prev_cpu usable by the task.
+	 */
+	cpu = pick_idle_cpu(p, prev_cpu, wake_flags);
+	if (cpu >= 0) {
+		if (can_direct_dispatch(cpu)) {
+			scx_bpf_dsq_insert_vtime(p, cpu_to_dsq(cpu),
+						 slice_ns, p->scx.dsq_vtime, 0);
+			__sync_fetch_and_add(&nr_kernel_dispatches, 1);
+		}
 		return cpu;
+	}
 
 	/*
 	 * If we couldn't find an idle CPU, in case of a sync wakeup
 	 * prioritize the waker's CPU.
 	 */
-	return is_wake_sync(wake_flags) ? bpf_get_smp_processor_id() : prev_cpu;
-
+	return prev_cpu;
 }
 
 /*
@@ -754,16 +613,48 @@ SEC("syscall")
 int rs_select_cpu(struct task_cpu_arg *input)
 {
 	struct task_struct *p;
-	int cpu;
+	int cpu = input->cpu;
 
 	p = bpf_task_from_pid(input->pid);
 	if (!p)
 		return -EINVAL;
 
+	/*
+	 * If the target CPU is the current one, treat it as idle when no
+	 * other tasks are queued.
+	 *
+	 * Since this function is invoked by the user-space scheduler,
+	 * which will release the CPU shortly, there is no need to migrate
+	 * the task elsewhere.
+	 */
+	if (cpu == bpf_get_smp_processor_id()) {
+		u64 nr_tasks = nr_running + nr_queued + nr_scheduled + 1;
+
+		if (nr_tasks < nr_online_cpus && !scx_bpf_dsq_nr_queued(cpu))
+			goto out_release;
+	}
+
 	bpf_rcu_read_lock();
-	cpu = pick_idle_cpu(p, input->cpu);
+	/*
+	 * Kernels that don't provide scx_bpf_select_cpu_and() only allow
+	 * to use the built-in idle CPU selection policy only from
+	 * ops.select_cpu() and opt.enqueue(), return any idle CPU usable
+	 * by the task in this case.
+	 */
+	if (!bpf_ksym_exists(scx_bpf_select_cpu_and)) {
+		if (!scx_bpf_test_and_clear_cpu_idle(cpu))
+			cpu = scx_bpf_pick_idle_cpu(p->cpus_ptr, 0);
+	} else {
+		/*
+		 * Set SCX_WAKE_TTWU, pretending to be a wakeup, to prioritize
+		 * faster CPU selection (we probably want to add an option to allow
+		 * the user-space scheduler to use this logic or not).
+		 */
+		cpu = pick_idle_cpu(p, cpu, SCX_WAKE_TTWU);
+	}
 	bpf_rcu_read_unlock();
 
+out_release:
 	bpf_task_release(p);
 
 	return cpu;
@@ -774,19 +665,21 @@ int rs_select_cpu(struct task_cpu_arg *input)
  * scheduler.
  */
 static void get_task_info(struct queued_task_ctx *task,
-			  const struct task_struct *p, u64 enq_flags)
+			  const struct task_struct *p,
+			  struct task_ctx *tctx, u64 enq_flags, s32 prev_cpu)
 {
-	struct task_ctx *tctx = try_lookup_task_ctx(p);
-
 	task->pid = p->pid;
-	task->cpu = scx_bpf_task_cpu(p);
+	task->cpu = prev_cpu;
 	task->nr_cpus_allowed = p->nr_cpus_allowed;
 	task->flags = enq_flags;
-	task->start_ts = tctx ? tctx->start_ts : 0;
-	task->stop_ts = tctx ? tctx->stop_ts : 0;
-	task->exec_runtime = tctx ? tctx->exec_runtime : 0;
+	task->start_ts = tctx->start_ts;
+	task->stop_ts = tctx->stop_ts;
+	task->exec_runtime = tctx->exec_runtime;
 	task->weight = p->scx.weight;
 	task->vtime = p->scx.dsq_vtime;
+	task->enq_cnt = ++tctx->enq_cnt;
+
+	bpf_core_read(&task->comm, sizeof(task->comm), &p->comm);
 }
 
 /*
@@ -807,6 +700,44 @@ static bool is_queued_wakeup(const struct task_struct *p, u64 enq_flags)
 	return !__COMPAT_is_enq_cpu_selected(enq_flags) && !scx_bpf_task_running(p);
 }
 
+/*
+ * Queue a task to the user-space scheduler.
+ */
+static void queue_task_to_userspace(struct task_struct *p, s32 prev_cpu, u64 enq_flags)
+{
+	struct queued_task_ctx *task;
+	struct task_ctx *tctx;
+
+	tctx = try_lookup_task_ctx(p);
+	if (!tctx)
+		return;
+
+	/*
+	 * Allocate a new entry in the ring buffer.
+	 *
+	 * If ring buffer is full, the user-space scheduler is congested,
+	 * so dispatch the task directly using the shared DSQ (the task
+	 * will be consumed by the first CPU available).
+	 */
+	task = bpf_ringbuf_reserve(&queued, sizeof(*task), 0);
+	if (!task) {
+		sched_congested(p);
+		scx_bpf_dsq_insert_vtime(p, SHARED_DSQ,
+					 slice_ns, p->scx.dsq_vtime, enq_flags);
+		__sync_fetch_and_add(&nr_kernel_dispatches, 1);
+		return;
+	}
+
+	/*
+	 * Collect task information and store them in the ring buffer that
+	 * will be consumed by the user-space scheduler.
+	 */
+	dbg_msg("enqueue: pid=%d (%s)", p->pid, p->comm);
+	get_task_info(task, p, tctx, enq_flags, prev_cpu);
+	bpf_ringbuf_submit(task, 0);
+	__sync_fetch_and_add(&nr_queued, 1);
+}
+
 /*
  * Task @p becomes ready to run. We can dispatch the task directly here if the
  * user-space scheduler is not required, or enqueue it to be processed by the
@@ -814,15 +745,18 @@ static bool is_queued_wakeup(const struct task_struct *p, u64 enq_flags)
  */
 void BPF_STRUCT_OPS(rustland_enqueue, struct task_struct *p, u64 enq_flags)
 {
-	struct queued_task_ctx *task;
-	s32 cpu = -EBUSY;
+	s32 prev_cpu = scx_bpf_task_cpu(p), cpu;
+	bool is_wakeup = is_queued_wakeup(p, enq_flags);
 
 	/*
-	 * Scheduler is dispatched directly in .dispatch() when needed, so
-	 * we can skip it here.
+	 * Insert the user-space scheduler to its dedicated DSQ, it will be
+	 * consumed from ops.dispatch() only when there's any pending
+	 * scheduling action to do.
 	 */
-	if (is_usersched_task(p))
-		return;
+	if (is_usersched_task(p)) {
+		scx_bpf_dsq_insert(p, SCHED_DSQ, slice_ns, enq_flags);
+		goto out_kick;
+	}
 
 	/*
 	 * Always dispatch per-CPU kthreads directly on their target CPU.
@@ -832,77 +766,73 @@ void BPF_STRUCT_OPS(rustland_enqueue, struct task_struct *p, u64 enq_flags)
 	 * (i.e., ksoftirqd/N, rcuop/N, etc.).
 	 */
 	if ((is_kthread(p) && p->nr_cpus_allowed == 1) || is_kswapd(p) || is_khugepaged(p)) {
-		cpu = scx_bpf_task_cpu(p);
-                scx_bpf_dsq_insert_vtime(p, cpu_to_dsq(cpu),
-					 SCX_SLICE_DFL, p->scx.dsq_vtime, enq_flags);
-		__sync_fetch_and_add(&nr_kernel_dispatches, 1);
-		return;
-	}
-
-	/*
-	 * Give the task a chance to be directly dispatched if
-	 * ops.select_cpu() was skipped.
-	 */
-	if (builtin_idle && is_queued_wakeup(p, enq_flags)) {
-		bool dispatched = false;
-
-		cpu = try_direct_dispatch(p, scx_bpf_task_cpu(p), enq_flags, &dispatched);
-		if (dispatched) {
-			scx_bpf_kick_cpu(cpu, SCX_KICK_IDLE);
-			return;
-		}
-	}
-
-	/*
-	 * Add tasks to the @queued list, they will be processed by the
-	 * user-space scheduler.
-	 *
-	 * If @queued list is full (user-space scheduler is congested) tasks
-	 * will be dispatched directly from the kernel (using the first CPU
-	 * available in this case).
-	 */
-	task = bpf_ringbuf_reserve(&queued, sizeof(*task), 0);
-	if (!task) {
-		sched_congested(p);
-		scx_bpf_dsq_insert_vtime(p, SHARED_DSQ, SCX_SLICE_DFL, p->scx.dsq_vtime, enq_flags);
+		scx_bpf_dsq_insert_vtime(p, cpu_to_dsq(prev_cpu),
+					 slice_ns, p->scx.dsq_vtime, enq_flags);
 		__sync_fetch_and_add(&nr_kernel_dispatches, 1);
 		goto out_kick;
 	}
-	get_task_info(task, p, enq_flags);
-	dbg_msg("enqueue: pid=%d (%s)", p->pid, p->comm);
-	bpf_ringbuf_submit(task, 0);
 
-	__sync_fetch_and_add(&nr_queued, 1);
-
-out_kick:
-	if (cpu < 0)
-		cpu = scx_bpf_task_cpu(p);
-	kick_task_cpu(p, cpu);
-}
-
-/*
- * Dispatch the user-space scheduler.
- */
-static void dispatch_user_scheduler(void)
-{
-	struct task_struct *p;
-
-	p = bpf_task_from_pid(usersched_pid);
-	if (!p) {
-		scx_bpf_error("Failed to find usersched task %d", usersched_pid);
-		return;
+	/*
+	 * If @builtin_idle is enabled, give the task a chance to be
+	 * directly dispatched only on a wakeup and only if
+	 * ops.select_cpu() was skipped, otherwise the task is always
+	 * queued to the user-space scheduler.
+	 */
+	if (!(builtin_idle && is_wakeup)) {
+		queue_task_to_userspace(p, prev_cpu, enq_flags);
+		goto out_kick;
 	}
 
 	/*
-	 * Assign an infinite time slice to the user-space scheduler, so
-	 * that it can completely drain all the pending tasks.
-	 *
-	 * The user-space scheduler will voluntarily yield the CPU upon
-	 * completion through BpfScheduler->notify_complete().
+	 * Try to find an idle CPU in the system, if all CPUs are busy
+	 * queue the task to the user-space scheduler.
 	 */
-	scx_bpf_dsq_insert(p, SCHED_DSQ, SCX_SLICE_INF, 0);
+	cpu = pick_idle_cpu(p, prev_cpu, 0);
+	if (cpu < 0) {
+		queue_task_to_userspace(p, prev_cpu, enq_flags);
+		goto out_kick;
+	}
 
-	bpf_task_release(p);
+	/*
+	 * Always force a CPU wakeup, so that the allocated CPU can be
+	 * released and go back idle even if the task isn't directly
+	 * dispatched.
+	 */
+	prev_cpu = cpu;
+	is_wakeup = true;
+
+	/*
+	 * Bounce the task to the user-space scheduler if we can't directly
+	 * dispatch to the selected CPU.
+	 */
+	if (!can_direct_dispatch(cpu)) {
+		queue_task_to_userspace(p, prev_cpu, enq_flags);
+		goto out_kick;
+	}
+
+	/*
+	 * We can race with a dequeue here and the selected idle CPU might
+	 * be not valid anymore, if the task affinity has changed.
+	 *
+	 * In this case just wakeup the picked CPU and ignore the enqueue,
+	 * another enqueue event for the same task will be received later.
+	 */
+	if (!bpf_cpumask_test_cpu(cpu, p->cpus_ptr))
+		goto out_kick;
+
+	/*
+	 * Directly dispatch the task to selected idle CPU (queued wakeup).
+	 */
+	scx_bpf_dsq_insert_vtime(p, cpu_to_dsq(cpu),
+				 slice_ns, p->scx.dsq_vtime, enq_flags);
+	__sync_fetch_and_add(&nr_kernel_dispatches, 1);
+
+out_kick:
+	/*
+	 * Wakeup the task's CPU if needed.
+	 */
+	if (is_wakeup)
+		scx_bpf_kick_cpu(prev_cpu, SCX_KICK_IDLE);
 }
 
 /*
@@ -934,19 +864,23 @@ static long handle_dispatched_task(struct bpf_dynptr *dynptr, void *context)
  */
 void BPF_STRUCT_OPS(rustland_dispatch, s32 cpu, struct task_struct *prev)
 {
-	/*
-	 * Fire up the user-space scheduler: it will run only if no other
-	 * task needs to run.
-	 */
-	if (test_and_clear_usersched_needed())
-		dispatch_user_scheduler();
-
 	/*
 	 * Consume all tasks from the @dispatched list and immediately
 	 * dispatch them on the target CPU decided by the user-space
 	 * scheduler.
 	 */
-	bpf_user_ringbuf_drain(&dispatched, handle_dispatched_task, NULL, BPF_RB_NO_WAKEUP);
+	s32 ret = bpf_user_ringbuf_drain(&dispatched,
+					 handle_dispatched_task, NULL, BPF_RB_NO_WAKEUP);
+	if (ret)
+		dbg_msg("User ringbuf drain error: %d", ret);
+
+	/*
+	 * Dispatch the user-space scheduler if there's any pending action
+	 * to do.
+	 */
+	if (usersched_has_pending_tasks() &&
+	    scx_bpf_dsq_move_to_local(SCHED_DSQ))
+		return;
 
 	/*
 	 * Consume a task from the per-CPU DSQ.
@@ -960,29 +894,17 @@ void BPF_STRUCT_OPS(rustland_dispatch, s32 cpu, struct task_struct *prev)
 	if (scx_bpf_dsq_move_to_local(SHARED_DSQ))
 		return;
 
-	/*
-	 * Lastly, consume and dispatch the user-space scheduler.
-	 */
-	if (scx_bpf_dsq_move_to_local(SCHED_DSQ))
-		return;
-
-	/*
-	 * If there are still pending task, notify the user-space scheduler
-	 * and prevent the CPU from going idle.
-	 */
-	if (usersched_has_pending_tasks()) {
-		set_usersched_needed();
-		scx_bpf_kick_cpu(cpu, SCX_KICK_IDLE);
-		return;
-	}
-
 	/*
 	 * If the current task expired its time slice and no other task
 	 * wants to run, simply replenish its time slice and let it run for
 	 * another round on the same CPU.
+	 *
+	 * In case of the user-space scheduler task, replenish its time
+	 * slice only if there're still pending scheduling actions to do.
 	 */
-	if (prev && is_queued(prev) && !is_usersched_task(prev))
-		prev->scx.slice = SCX_SLICE_DFL;
+	if (prev && is_queued(prev) &&
+	    (!is_usersched_task(prev) || usersched_has_pending_tasks()))
+		prev->scx.slice = slice_ns;
 }
 
 void BPF_STRUCT_OPS(rustland_runnable, struct task_struct *p, u64 enq_flags)
@@ -1053,70 +975,13 @@ void BPF_STRUCT_OPS(rustland_stopping, struct task_struct *p, bool runnable)
 	tctx->exec_runtime += now - tctx->start_ts;
 }
 
-/*
- * A CPU is taken away from the scheduler, preempting the current task by
- * another one running in a higher priority sched_class.
- */
-void BPF_STRUCT_OPS(rustland_cpu_release, s32 cpu,
-				struct scx_cpu_release_args *args)
-{
-	struct task_struct *p = args->task;
-	/*
-	 * If the interrupted task is the user-space scheduler make sure to
-	 * re-schedule it immediately.
-	 */
-	dbg_msg("cpu preemption: pid=%d (%s)", p->pid, p->comm);
-	if (is_usersched_task(p))
-		set_usersched_needed();
-}
-
 /*
  * A task joins the sched_ext scheduler.
  */
 void BPF_STRUCT_OPS(rustland_enable, struct task_struct *p)
 {
 	p->scx.dsq_vtime = 0;
-	p->scx.slice = SCX_SLICE_DFL;
-}
-
-/*
- * A new task @p is being created.
- *
- * Allocate and initialize all the internal structures for the task (this
- * function is allowed to block, so it can be used to preallocate memory).
- */
-s32 BPF_STRUCT_OPS(rustland_init_task, struct task_struct *p,
-		   struct scx_init_task_args *args)
-{
-	struct task_ctx *tctx;
-	struct bpf_cpumask *cpumask;
-
-	tctx = bpf_task_storage_get(&task_ctx_stor, p, 0,
-				    BPF_LOCAL_STORAGE_GET_F_CREATE);
-	if (!tctx)
-		return -ENOMEM;
-
-	/*
-	 * Create task's L2 cache cpumask.
-	 */
-	cpumask = bpf_cpumask_create();
-	if (!cpumask)
-		return -ENOMEM;
-	cpumask = bpf_kptr_xchg(&tctx->l2_cpumask, cpumask);
-	if (cpumask)
-		bpf_cpumask_release(cpumask);
-
-	/*
-	 * Create task's L3 cache cpumask.
-	 */
-	cpumask = bpf_cpumask_create();
-	if (!cpumask)
-		return -ENOMEM;
-	cpumask = bpf_kptr_xchg(&tctx->l3_cpumask, cpumask);
-	if (cpumask)
-		bpf_cpumask_release(cpumask);
-
-	return 0;
+	p->scx.slice = slice_ns;
 }
 
 /*
@@ -1141,12 +1006,8 @@ static int usersched_timer_fn(void *map, int *key, struct bpf_timer *timer)
 		bpf_rcu_read_lock();
 		p = bpf_task_from_pid(usersched_pid);
 		if (p) {
-			s32 cpu;
-
 			set_usersched_needed();
-			cpu = scx_bpf_pick_idle_cpu(p->cpus_ptr, 0);
-			if (cpu >= 0)
-				scx_bpf_kick_cpu(cpu, SCX_KICK_IDLE);
+			scx_bpf_kick_cpu(scx_bpf_task_cpu(p), SCX_KICK_IDLE);
 			bpf_task_release(p);
 		}
 		bpf_rcu_read_unlock();
@@ -1248,62 +1109,23 @@ static int dsq_init(void)
 	return 0;
 }
 
-static int init_cpumask(struct bpf_cpumask **cpumask)
+/*
+ * A new task @p is being created.
+ *
+ * Allocate and initialize all the internal structures for the task (this
+ * function is allowed to block, so it can be used to preallocate memory).
+ */
+s32 BPF_STRUCT_OPS(rustland_init_task, struct task_struct *p,
+		   struct scx_init_task_args *args)
 {
-	struct bpf_cpumask *mask;
-	int err = 0;
+	struct task_ctx *tctx;
 
-	/*
-	 * Do nothing if the mask is already initialized.
-	 */
-	mask = *cpumask;
-	if (mask)
-		return 0;
-	/*
-	 * Create the CPU mask.
-	 */
-	err = calloc_cpumask(cpumask);
-	if (!err)
-		mask = *cpumask;
-	if (!mask)
-		err = -ENOMEM;
+	tctx = bpf_task_storage_get(&task_ctx_stor, p, 0,
+				    BPF_LOCAL_STORAGE_GET_F_CREATE);
+	if (!tctx)
+		return -ENOMEM;
 
-	return err;
-}
-
-SEC("syscall")
-int enable_sibling_cpu(struct domain_arg *input)
-{
-	struct cpu_ctx *cctx;
-	struct bpf_cpumask *mask, **pmask;
-	int err = 0;
-
-	cctx = try_lookup_cpu_ctx(input->cpu_id);
-	if (!cctx)
-		return -ENOENT;
-
-	/* Make sure the target CPU mask is initialized */
-	switch (input->lvl_id) {
-	case 2:
-		pmask = &cctx->l2_cpumask;
-		break;
-	case 3:
-		pmask = &cctx->l3_cpumask;
-		break;
-	default:
-		return -EINVAL;
-	}
-	err = init_cpumask(pmask);
-	if (err)
-		return err;
-
-	bpf_rcu_read_lock();
-	mask = *pmask;
-	if (mask)
-		bpf_cpumask_set_cpu(input->sibling_cpu_id, mask);
-	bpf_rcu_read_unlock();
-
-	return err;
+	return 0;
 }
 
 /*
@@ -1348,7 +1170,6 @@ SCX_OPS_DEFINE(rustland,
 	       .runnable		= (void *)rustland_runnable,
 	       .running			= (void *)rustland_running,
 	       .stopping		= (void *)rustland_stopping,
-	       .cpu_release		= (void *)rustland_cpu_release,
 	       .enable			= (void *)rustland_enable,
 	       .init_task		= (void *)rustland_init_task,
 	       .init			= (void *)rustland_init,
