@@ -12,47 +12,53 @@
 
 #include <stdbool.h>
 
-#define __COMPAT_chaos_scx_bpf_dsq_move_set_slice(it__iter, slice)				\
-	(bpf_ksym_exists(scx_bpf_dsq_move_set_slice) ?						\
-	  scx_bpf_dsq_move_set_slice((it__iter), (slice)) :					\
-	  scx_bpf_dispatch_from_dsq_set_slice___compat((it__iter), (slice)))
+#define __COMPAT_chaos_scx_bpf_dsq_move_set_slice(it__iter, slice)        \
+	(bpf_ksym_exists(scx_bpf_dsq_move_set_slice) ?                    \
+		 scx_bpf_dsq_move_set_slice((it__iter), (slice)) :        \
+		 scx_bpf_dispatch_from_dsq_set_slice___compat((it__iter), \
+							      (slice)))
 
-#define __COMPAT_chaos_scx_bpf_dsq_move(it__iter, p, dsq_id, enq_flags)				\
-	(bpf_ksym_exists(scx_bpf_dsq_move) ?							\
-	  scx_bpf_dsq_move((it__iter), (p), (dsq_id), (enq_flags)) :				\
-	  scx_bpf_dispatch_from_dsq___compat((it__iter), (p), (dsq_id), (enq_flags)))
+#define __COMPAT_chaos_scx_bpf_dsq_move(it__iter, p, dsq_id, enq_flags)        \
+	(bpf_ksym_exists(scx_bpf_dsq_move) ?                                   \
+		 scx_bpf_dsq_move((it__iter), (p), (dsq_id), (enq_flags)) :    \
+		 scx_bpf_dispatch_from_dsq___compat((it__iter), (p), (dsq_id), \
+						    (enq_flags)))
 
-#define __COMPAT_chaos_scx_bpf_dsq_move_set_vtime(it__iter, vtime)				\
-	(bpf_ksym_exists(scx_bpf_dsq_move_set_vtime) ?						\
-	  scx_bpf_dsq_move_set_vtime((it__iter), (vtime)) :					\
-	  scx_bpf_dispatch_from_dsq_set_vtime___compat((it__iter), (vtime)))
+#define __COMPAT_chaos_scx_bpf_dsq_move_set_vtime(it__iter, vtime)        \
+	(bpf_ksym_exists(scx_bpf_dsq_move_set_vtime) ?                    \
+		 scx_bpf_dsq_move_set_vtime((it__iter), (vtime)) :        \
+		 scx_bpf_dispatch_from_dsq_set_vtime___compat((it__iter), \
+							      (vtime)))
 
-#define __COMPAT_chaos_scx_bpf_dsq_move_vtime(it__iter, p, dsq_id, enq_flags)			\
-	(bpf_ksym_exists(scx_bpf_dsq_move_vtime) ?						\
-	  scx_bpf_dsq_move_vtime((it__iter), (p), (dsq_id), (enq_flags)) :			\
-	  scx_bpf_dispatch_vtime_from_dsq___compat((it__iter), (p), (dsq_id), (enq_flags)))
+#define __COMPAT_chaos_scx_bpf_dsq_move_vtime(it__iter, p, dsq_id, enq_flags) \
+	(bpf_ksym_exists(scx_bpf_dsq_move_vtime) ?                            \
+		 scx_bpf_dsq_move_vtime((it__iter), (p), (dsq_id),            \
+					(enq_flags)) :                        \
+		 scx_bpf_dispatch_vtime_from_dsq___compat(                    \
+			 (it__iter), (p), (dsq_id), (enq_flags)))
 
-const volatile int ppid_targeting_ppid = 1;
-const volatile bool ppid_targeting_inclusive = false; /* include ppid_targeting_ppid in chaos */
+const volatile int  ppid_targeting_ppid = 1;
+const volatile bool ppid_targeting_inclusive =
+	false; /* include ppid_targeting_ppid in chaos */
 
-const volatile u64 chaos_timer_check_queues_min_ns = 500000;
-const volatile u64 chaos_timer_check_queues_max_ns = 2000000;
+const volatile u64 chaos_timer_check_queues_min_ns   = 500000;
+const volatile u64 chaos_timer_check_queues_max_ns   = 2000000;
 const volatile u64 chaos_timer_check_queues_slack_ns = 2500000;
 
 const volatile u32 trait_delay_freq_frac32[CHAOS_TRAIT_MAX];
 
-const volatile u64 random_delays_min_ns = 1; /* for veristat */
-const volatile u64 random_delays_max_ns = 2; /* for veristat */
+const volatile u64 random_delays_min_ns	     = 1; /* for veristat */
+const volatile u64 random_delays_max_ns	     = 2; /* for veristat */
 
-const volatile u32 cpu_freq_min = 0;
-const volatile u32 cpu_freq_max = SCX_CPUPERF_ONE;
+const volatile u32 cpu_freq_min		     = 0;
+const volatile u32 cpu_freq_max		     = SCX_CPUPERF_ONE;
 
-const volatile u32 degradation_freq_frac32 = 1;
-const volatile u64 degradation_frac7 = 0;
+const volatile u32 degradation_freq_frac32   = 1;
+const volatile u64 degradation_frac7	     = 0;
 
 const volatile u32 kprobe_delays_freq_frac32 = 1;
-const volatile u64 kprobe_delays_min_ns = 1;
-const volatile u64 kprobe_delays_max_ns = 2;
+const volatile u64 kprobe_delays_min_ns	     = 1;
+const volatile u64 kprobe_delays_max_ns	     = 2;
 
 #define MIN(x, y) ((x) < (y) ? (x) : (y))
 #define MAX(x, y) ((x) > (y) ? (x) : (y))
@@ -81,11 +87,12 @@ struct {
 	__uint(max_entries, CHAOS_NR_STATS);
 	__type(key, u32);
 	__type(value, u64);
-} chaos_stats SEC(".maps");
+} chaos_stats	       SEC(".maps");
 
 struct chaos_task_ctx *lookup_create_chaos_task_ctx(struct task_struct *p)
 {
-	return bpf_task_storage_get(&chaos_task_ctxs, p, NULL, BPF_LOCAL_STORAGE_GET_F_CREATE);
+	return bpf_task_storage_get(&chaos_task_ctxs, p, NULL,
+				    BPF_LOCAL_STORAGE_GET_F_CREATE);
 }
 
 static __always_inline u64 chaos_get_prandom_u64_limit(u64 s)
@@ -108,11 +115,12 @@ static __always_inline u64 chaos_get_prandom_u64_limit(u64 s)
 
 	if (m_low < s) {
 		t = ((u64)(-s)) % s;
-		bpf_repeat(CHAOS_MAX_RAND_ATTEMPTS) {
+		bpf_repeat(CHAOS_MAX_RAND_ATTEMPTS)
+		{
 			if (m_low >= t)
 				break;
 
-			x = get_prandom_u64();
+			x      = get_prandom_u64();
 			m_high = ((x >> 32) * (s >> 32)) +
 				 (((x & 0xFFFFFFFF) * (s >> 32)) >> 32) +
 				 (((x >> 32) * (s & 0xFFFFFFFF)) >> 32);
@@ -135,7 +143,8 @@ static __always_inline void chaos_stat_inc(enum chaos_stat_idx stat)
 		(*cnt_p)++;
 }
 
-static __always_inline enum chaos_trait_kind choose_chaos(struct chaos_task_ctx *taskc)
+static __always_inline enum chaos_trait_kind
+choose_chaos(struct chaos_task_ctx *taskc)
 {
 	if (taskc->match & CHAOS_MATCH_EXCLUDED) {
 		chaos_stat_inc(CHAOS_STAT_CHAOS_EXCLUDED);
@@ -144,7 +153,7 @@ static __always_inline enum chaos_trait_kind choose_chaos(struct chaos_task_ctx
 
 	u32 roll = bpf_get_prandom_u32();
 
-	#pragma unroll
+#pragma unroll
 	for (int i = 0; i < CHAOS_TRAIT_MAX; ++i) {
 		if (roll <= trait_delay_freq_frac32[i])
 			return i;
@@ -154,10 +163,11 @@ static __always_inline enum chaos_trait_kind choose_chaos(struct chaos_task_ctx
 	return -1;
 }
 
-static __always_inline bool chaos_trait_skips_select_cpu(struct chaos_task_ctx *taskc)
+static __always_inline bool
+chaos_trait_skips_select_cpu(struct chaos_task_ctx *taskc)
 {
-	return taskc->next_trait == CHAOS_TRAIT_RANDOM_DELAYS || 
-		taskc->next_trait == CHAOS_TRAIT_KPROBE_RANDOM_DELAYS;
+	return taskc->next_trait == CHAOS_TRAIT_RANDOM_DELAYS ||
+	       taskc->next_trait == CHAOS_TRAIT_KPROBE_RANDOM_DELAYS;
 }
 
 static __always_inline u64 get_cpu_delay_dsq(int cpu_idx)
@@ -175,11 +185,11 @@ static __always_inline u64 get_cpu_delay_dsq(int cpu_idx)
 static __always_inline s32 calculate_chaos_match(struct task_struct *p)
 {
 	struct chaos_task_ctx *taskc;
-	struct task_struct *p2;
-	enum chaos_match flags = 0;
-	int found_parent = 0;
-	int ret = 0;
-	int pid;
+	struct task_struct    *p2;
+	enum chaos_match       flags	    = 0;
+	int		       found_parent = 0;
+	int		       ret	    = 0;
+	int		       pid;
 
 	if (!(taskc = lookup_create_chaos_task_ctx(p))) {
 		scx_bpf_error("couldn't create task context");
@@ -196,7 +206,8 @@ static __always_inline s32 calculate_chaos_match(struct task_struct *p)
 	// no need for the path-to-root walk, this is the task
 	if (ppid_targeting_ppid == p->pid) {
 		taskc->match |= ppid_targeting_inclusive ?
-			CHAOS_MATCH_HAS_PARENT : CHAOS_MATCH_EXCLUDED;
+					CHAOS_MATCH_HAS_PARENT :
+					CHAOS_MATCH_EXCLUDED;
 		return 0;
 	}
 
@@ -207,7 +218,8 @@ static __always_inline s32 calculate_chaos_match(struct task_struct *p)
 	}
 
 	// walk the real_parent path-to-root to check for the HAS_PARENT match
-	bpf_repeat(CHAOS_NUM_PPIDS_CHECK) {
+	bpf_repeat(CHAOS_NUM_PPIDS_CHECK)
+	{
 		p2 = bpf_task_from_pid(pid);
 		if (!p2)
 			break;
@@ -262,7 +274,8 @@ static __always_inline s32 calculate_chaos_match(struct task_struct *p)
 	if (!p->real_parent || !(pid = p->real_parent->pid))
 		return 0;
 
-	bpf_repeat(CHAOS_NUM_PPIDS_CHECK) {
+	bpf_repeat(CHAOS_NUM_PPIDS_CHECK)
+	{
 		p2 = bpf_task_from_pid(pid);
 		if (!p2)
 			break;
@@ -293,10 +306,12 @@ out:
 	return ret;
 }
 
-__weak s32 enqueue_random_delay(struct task_struct *p __arg_trusted, u64 enq_flags,
-				struct chaos_task_ctx *taskc __arg_nonnull, u64 min_ns, u64 max_ns)
+__weak s32 enqueue_random_delay(struct task_struct *p	     __arg_trusted,
+				u64			     enq_flags,
+				struct chaos_task_ctx *taskc __arg_nonnull,
+				u64 min_ns, u64 max_ns)
 {
-	u64 vtime = chaos_get_uniform_u64(min_ns, max_ns);
+	u64 vtime = bpf_ktime_get_ns() + chaos_get_uniform_u64(min_ns, max_ns);
 	scx_bpf_dsq_insert_vtime(p, get_cpu_delay_dsq(-1), 0, vtime, enq_flags);
 	return true;
 }
@@ -308,17 +323,13 @@ __weak s32 enqueue_chaotic(struct task_struct *p __arg_trusted, u64 enq_flags,
 
 	switch (taskc->next_trait) {
 	case CHAOS_TRAIT_KPROBE_RANDOM_DELAYS:
-		out = enqueue_random_delay(p,
-					   enq_flags,
-					   taskc,
+		out = enqueue_random_delay(p, enq_flags, taskc,
 					   kprobe_delays_min_ns,
 					   kprobe_delays_max_ns);
 		chaos_stat_inc(CHAOS_STAT_KPROBE_RANDOM_DELAYS);
 		break;
 	case CHAOS_TRAIT_RANDOM_DELAYS:
-		out = enqueue_random_delay(p,
-					   enq_flags,
-					   taskc,
+		out = enqueue_random_delay(p, enq_flags, taskc,
 					   random_delays_min_ns,
 					   random_delays_max_ns);
 		chaos_stat_inc(CHAOS_STAT_TRAIT_RANDOM_DELAYS);
@@ -348,9 +359,9 @@ __weak s32 enqueue_chaotic(struct task_struct *p __arg_trusted, u64 enq_flags,
 __weak u64 check_dsq_times(int cpu_idx)
 {
 	struct task_struct *p;
-	u64 next_trigger_time = 0;
-	u64 now = bpf_ktime_get_ns();
-	bool has_kicked = false;
+	u64		    next_trigger_time = 0;
+	u64		    now		      = bpf_ktime_get_ns();
+	bool		    has_kicked	      = false;
 
 	bpf_rcu_read_lock();
 	bpf_for_each(scx_dsq, p, get_cpu_delay_dsq(cpu_idx), 0) {
@@ -358,7 +369,9 @@ __weak u64 check_dsq_times(int cpu_idx)
 		if (!p)
 			break;
 
-		if (!has_kicked && p->scx.dsq_vtime < now - chaos_timer_check_queues_slack_ns) {
+		if (!has_kicked &&
+		    p->scx.dsq_vtime <
+			    now - chaos_timer_check_queues_slack_ns) {
 			has_kicked = true;
 			scx_bpf_kick_cpu(cpu_idx, SCX_KICK_PREEMPT);
 			chaos_stat_inc(CHAOS_STAT_TIMER_KICKS);
@@ -379,39 +392,45 @@ __weak u64 check_dsq_times(int cpu_idx)
 	return next_trigger_time;
 }
 
-static int chaos_timer_check_queues_callback(void *map, int key, struct timer_wrapper *timerw)
+static int chaos_timer_check_queues_callback(void *map, int key,
+					     struct timer_wrapper *timerw)
 {
-	u64 started_at = bpf_ktime_get_ns();
+	u64 started_at	      = bpf_ktime_get_ns();
 	u64 next_trigger_time = 0;
 	u64 this_next_trigger_time;
 	int cpu_idx;
 
-	bpf_for(cpu_idx, 0, topo_config.nr_cpus) {
+	bpf_for(cpu_idx, 0, topo_config.nr_cpus)
+	{
 		this_next_trigger_time = check_dsq_times(cpu_idx);
-		next_trigger_time = MAX(next_trigger_time, this_next_trigger_time);
+		next_trigger_time =
+			MAX(next_trigger_time, this_next_trigger_time);
 	}
 
 	if (next_trigger_time == 0) {
-		bpf_timer_start(&timerw->timer, chaos_timer_check_queues_max_ns, 0);
+		bpf_timer_start(&timerw->timer, chaos_timer_check_queues_max_ns,
+				0);
 		return 0;
 	}
 
-	next_trigger_time = MAX(next_trigger_time, started_at + chaos_timer_check_queues_min_ns);
-	next_trigger_time = MIN(next_trigger_time, started_at + chaos_timer_check_queues_max_ns);
+	next_trigger_time = MAX(next_trigger_time,
+				started_at + chaos_timer_check_queues_min_ns);
+	next_trigger_time = MIN(next_trigger_time,
+				started_at + chaos_timer_check_queues_max_ns);
 
 	bpf_timer_start(&timerw->timer, next_trigger_time, BPF_F_TIMER_ABS);
 	return 0;
 }
 
-
 s32 BPF_STRUCT_OPS_SLEEPABLE(chaos_init)
 {
 	struct timer_wrapper *timerw;
-	struct llc_ctx *llcx;
-	struct cpu_ctx *cpuc;
-	int timer_id, ret, i;
+	struct llc_ctx	     *llcx;
+	struct cpu_ctx	     *cpuc;
+	int		      timer_id, ret, i;
 
-	bpf_for(i, 0, topo_config.nr_cpus) {
+	bpf_for(i, 0, topo_config.nr_cpus)
+	{
 		if (!(cpuc = lookup_cpu_ctx(i)) ||
 		    !(llcx = lookup_llc_ctx(cpuc->llc_id)))
 			return -EINVAL;
@@ -422,7 +441,7 @@ s32 BPF_STRUCT_OPS_SLEEPABLE(chaos_init)
 	}
 
 	timer_id = CHAOS_TIMER_CHECK_QUEUES;
-	timerw = bpf_map_lookup_elem(&chaos_timers, &timer_id);
+	timerw	 = bpf_map_lookup_elem(&chaos_timers, &timer_id);
 	if (!timerw)
 		return -1;
 
@@ -432,11 +451,13 @@ s32 BPF_STRUCT_OPS_SLEEPABLE(chaos_init)
 	if (ret)
 		return -1;
 
-	ret = bpf_timer_set_callback(&timerw->timer, &chaos_timer_check_queues_callback);
+	ret = bpf_timer_set_callback(&timerw->timer,
+				     &chaos_timer_check_queues_callback);
 	if (ret)
 		return -1;
 
-	ret = bpf_timer_start(&timerw->timer, chaos_timer_check_queues_max_ns, 0);
+	ret = bpf_timer_start(&timerw->timer, chaos_timer_check_queues_max_ns,
+			      0);
 	if (ret)
 		return -1;
 
@@ -447,32 +468,40 @@ s32 BPF_STRUCT_OPS_SLEEPABLE(chaos_init)
  * Cleanup any outstanding work left over by the p2dq promise when we do not
  * plan to complete it.
  */
-static __always_inline void destroy_p2dq_enqueue_promise(struct enqueue_promise *pro)
+static __always_inline void
+destroy_p2dq_enqueue_promise(struct enqueue_promise *pro)
 {
 	// If the idle bit of a CPU has already been cleared but we don't plan
 	// to execute a task on it we should kick the CPU. If the CPU goes to
 	// sleep again it will reset the kernel managed idle state.
-	if (pro->has_cleared_idle)
+	if (enqueue_promise_test_flag(pro, ENQUEUE_PROMISE_F_HAS_CLEARED_IDLE))
 		scx_bpf_kick_cpu(pro->cpu, SCX_KICK_IDLE);
 }
 
-static __always_inline void complete_p2dq_enqueue_move(struct enqueue_promise *pro,
-						       struct bpf_iter_scx_dsq *it__iter,
-						       struct task_struct *p)
+static __always_inline void
+complete_p2dq_enqueue_move(struct enqueue_promise  *pro,
+			   struct bpf_iter_scx_dsq *it__iter,
+			   struct task_struct	   *p)
 {
 	switch (pro->kind) {
 	case P2DQ_ENQUEUE_PROMISE_COMPLETE:
-		scx_bpf_error("chaos: delayed async_p2dq_enqueue returned COMPLETE"
-			      " after a task was placed in the delay dsq!");
+		scx_bpf_error(
+			"chaos: delayed async_p2dq_enqueue returned COMPLETE"
+			" after a task was placed in the delay dsq!");
 		break;
 	case P2DQ_ENQUEUE_PROMISE_FIFO:
-		__COMPAT_chaos_scx_bpf_dsq_move_set_slice(it__iter, *MEMBER_VPTR(pro->fifo, .slice_ns));
-		__COMPAT_chaos_scx_bpf_dsq_move(it__iter, p, pro->fifo.dsq_id, pro->fifo.enq_flags);
+		__COMPAT_chaos_scx_bpf_dsq_move_set_slice(
+			it__iter, *MEMBER_VPTR(pro->fifo, .slice_ns));
+		__COMPAT_chaos_scx_bpf_dsq_move(it__iter, p, pro->fifo.dsq_id,
+						pro->fifo.enq_flags);
 		break;
 	case P2DQ_ENQUEUE_PROMISE_VTIME:
-		__COMPAT_chaos_scx_bpf_dsq_move_set_slice(it__iter, pro->vtime.slice_ns);
-		__COMPAT_chaos_scx_bpf_dsq_move_set_vtime(it__iter, pro->vtime.vtime);
-		__COMPAT_chaos_scx_bpf_dsq_move_vtime(it__iter, p, pro->vtime.dsq_id, pro->vtime.enq_flags);
+		__COMPAT_chaos_scx_bpf_dsq_move_set_slice(it__iter,
+							  pro->vtime.slice_ns);
+		__COMPAT_chaos_scx_bpf_dsq_move_set_vtime(it__iter,
+							  pro->vtime.vtime);
+		__COMPAT_chaos_scx_bpf_dsq_move_vtime(
+			it__iter, p, pro->vtime.dsq_id, pro->vtime.enq_flags);
 		break;
 	case P2DQ_ENQUEUE_PROMISE_ATQ_FIFO:
 	case P2DQ_ENQUEUE_PROMISE_ATQ_VTIME:
@@ -483,15 +512,15 @@ static __always_inline void complete_p2dq_enqueue_move(struct enqueue_promise *p
 		break;
 	}
 
-	if (pro->kick_idle)
+	if (enqueue_promise_test_flag(pro, ENQUEUE_PROMISE_F_KICK_IDLE))
 		scx_bpf_kick_cpu(pro->cpu, SCX_KICK_IDLE);
 
 	pro->kind = P2DQ_ENQUEUE_PROMISE_COMPLETE;
 }
 
 __weak int async_p2dq_enqueue_weak(struct enqueue_promise *ret __arg_nonnull,
-				   struct task_struct *p __arg_trusted,
-				   u64 enq_flags)
+				   struct task_struct *p       __arg_trusted,
+				   u64			       enq_flags)
 {
 	async_p2dq_enqueue(ret, p, enq_flags);
 	return 0;
@@ -501,8 +530,8 @@ void BPF_STRUCT_OPS(chaos_dispatch, s32 cpu, struct task_struct *prev)
 {
 	struct enqueue_promise promise;
 	struct chaos_task_ctx *taskc;
-	struct task_struct *p;
-	u64 now = bpf_ktime_get_ns();
+	struct task_struct    *p;
+	u64		       now = bpf_ktime_get_ns();
 
 	bpf_for_each(scx_dsq, p, get_cpu_delay_dsq(-1), 0) {
 		p = bpf_task_from_pid(p->pid);
@@ -531,7 +560,8 @@ void BPF_STRUCT_OPS(chaos_dispatch, s32 cpu, struct task_struct *prev)
 	return p2dq_dispatch_impl(cpu, prev);
 }
 
-void BPF_STRUCT_OPS(chaos_enqueue, struct task_struct *p __arg_trusted, u64 enq_flags)
+void BPF_STRUCT_OPS(chaos_enqueue, struct task_struct *p __arg_trusted,
+		    u64 enq_flags)
 {
 	struct enqueue_promise promise;
 	struct chaos_task_ctx *taskc;
@@ -551,7 +581,7 @@ void BPF_STRUCT_OPS(chaos_enqueue, struct task_struct *p __arg_trusted, u64 enq_
 		goto cleanup;
 
 	if ((taskc->next_trait == CHAOS_TRAIT_RANDOM_DELAYS ||
-		taskc->next_trait == CHAOS_TRAIT_KPROBE_RANDOM_DELAYS) &&
+	     taskc->next_trait == CHAOS_TRAIT_KPROBE_RANDOM_DELAYS) &&
 	    enqueue_chaotic(p, enq_flags, taskc))
 		goto cleanup;
 
@@ -559,13 +589,20 @@ void BPF_STRUCT_OPS(chaos_enqueue, struct task_struct *p __arg_trusted, u64 enq_
 	// direct dispatch in some situations.
 	if (taskc->next_trait == CHAOS_TRAIT_DEGRADATION) {
 		if (promise.kind == P2DQ_ENQUEUE_PROMISE_FIFO) {
-			promise.fifo.slice_ns = ((degradation_frac7 << 7) * promise.fifo.slice_ns) >> 7;
-			dbg("CHAOS[degradation][%d] slice_ns: %llu", p, promise.fifo.slice_ns);
+			promise.fifo.slice_ns = ((degradation_frac7 << 7) *
+						 promise.fifo.slice_ns) >>
+						7;
+			dbg("CHAOS[degradation][%d] slice_ns: %llu", p,
+			    promise.fifo.slice_ns);
 			chaos_stat_inc(CHAOS_STAT_TRAIT_DEGRADATION);
 		}
 		if (promise.kind == P2DQ_ENQUEUE_PROMISE_VTIME) {
-			promise.vtime.vtime += ((degradation_frac7 << 7) * promise.vtime.slice_ns) >> 7;
-			promise.vtime.slice_ns = ((degradation_frac7 << 7) * promise.vtime.slice_ns) >> 7;
+			promise.vtime.vtime += ((degradation_frac7 << 7) *
+						promise.vtime.slice_ns) >>
+					       7;
+			promise.vtime.slice_ns = ((degradation_frac7 << 7) *
+						  promise.vtime.slice_ns) >>
+						 7;
 			dbg("CHAOS[degradation][%d] vtime: %llu slice_ns: %llu",
 			    p, promise.vtime.vtime, promise.vtime.slice_ns);
 			chaos_stat_inc(CHAOS_STAT_TRAIT_DEGRADATION);
@@ -586,7 +623,7 @@ void BPF_STRUCT_OPS(chaos_runnable, struct task_struct *p, u64 enq_flags)
 		return;
 
 	if (wakee_ctx->pending_trait) {
-		wakee_ctx->next_trait = wakee_ctx->pending_trait;
+		wakee_ctx->next_trait	 = wakee_ctx->pending_trait;
 		wakee_ctx->pending_trait = 0;
 		return;
 	}
@@ -620,7 +657,8 @@ void BPF_STRUCT_OPS(chaos_running, struct task_struct *p)
 	}
 }
 
-s32 BPF_STRUCT_OPS(chaos_select_cpu, struct task_struct *p, s32 prev_cpu, u64 wake_flags)
+s32 BPF_STRUCT_OPS(chaos_select_cpu, struct task_struct *p, s32 prev_cpu,
+		   u64 wake_flags)
 {
 	struct chaos_task_ctx *wakee_ctx;
 	if (!(wakee_ctx = lookup_create_chaos_task_ctx(p)))
@@ -658,6 +696,30 @@ s32 BPF_STRUCT_OPS_SLEEPABLE(chaos_init_task, struct task_struct *p,
 	return 0;
 }
 
+SEC("kprobe/generic")
+int generic(struct pt_regs *ctx)
+{
+	struct task_struct    *p;
+	struct chaos_task_ctx *taskc;
+
+	p = (struct task_struct *)bpf_get_current_task_btf();
+	if (!p)
+		return -EINVAL;
+
+	if (!(taskc = lookup_create_chaos_task_ctx(p)))
+		return -EINVAL;
+
+	u32 roll = bpf_get_prandom_u32();
+	if (roll <= kprobe_delays_freq_frac32) {
+		taskc->pending_trait = CHAOS_TRAIT_KPROBE_RANDOM_DELAYS;
+		dbg("GENERIC: setting pending_trait to RANDOM_DELAYS - task[%d]",
+		    p->pid);
+	}
+
+	return 0;
+}
+
+// clang-format off
 SCX_OPS_DEFINE(chaos,
 	       .dispatch		= (void *)chaos_dispatch,
 	       .enqueue			= (void *)chaos_enqueue,
@@ -677,25 +739,4 @@ SCX_OPS_DEFINE(chaos,
 
 	       .timeout_ms		= 30000,
 	       .name			= "chaos");
-
-SEC("kprobe/generic")
-int generic(struct pt_regs *ctx)
-{
-	struct task_struct *p;
-	struct chaos_task_ctx *taskc;
-
-	p = (struct task_struct *)bpf_get_current_task_btf();
-	if (!p)
-		return -EINVAL;
-
-	if (!(taskc = lookup_create_chaos_task_ctx(p)))
-		return -EINVAL;
-
-	u32 roll = bpf_get_prandom_u32();
-	if (roll <= kprobe_delays_freq_frac32) {
-		taskc->pending_trait = CHAOS_TRAIT_KPROBE_RANDOM_DELAYS;
-		dbg("GENERIC: setting pending_trait to RANDOM_DELAYS - task[%d]", p->pid);
-	}
-
-	return 0;
-}
+// clang-format on
