@@ -22,15 +22,12 @@ use scx_utils::scx_ops_load;
 use scx_utils::scx_ops_open;
 use scx_utils::uei_exited;
 use scx_utils::uei_report;
-
-use scx_utils::Core;
-use scx_utils::Llc;
 use scx_utils::Topology;
-use scx_utils::NR_CPU_IDS;
 
-use scx_p2dq::bpf_intf::consts_STATIC_ALLOC_PAGES_GRANULARITY;
+use libbpf_rs::skel::Skel;
+use scx_arena::ArenaLib;
 use scx_p2dq::types;
-use std::ffi::c_ulong;
+use scx_utils::NR_CPU_IDS;
 
 use anyhow::bail;
 use anyhow::Context;
@@ -42,7 +39,6 @@ use libbpf_rs::AsRawLibbpf;
 use libbpf_rs::Link;
 use libbpf_rs::MapCore as _;
 use libbpf_rs::OpenObject;
-use libbpf_rs::ProgramInput;
 use log::debug;
 use log::info;
 use nix::unistd::Pid;
@@ -255,128 +251,6 @@ impl Drop for Scheduler {
 }
 
 impl Builder<'_> {
-    fn setup_arenas(&self, skel: &mut BpfSkel) -> Result<()> {
-        // Allocate the arena memory from the BPF side so userspace initializes it before starting
-        // the scheduler. Despite the function call's name this is neither a test nor a test run,
-        // it's the recommended way of executing SEC("syscall") probes.
-        let mut args = types::arena_init_args {
-            static_pages: consts_STATIC_ALLOC_PAGES_GRANULARITY as c_ulong,
-            task_ctx_size: std::mem::size_of::<types::task_p2dq>() as c_ulong,
-        };
-
-        let input = ProgramInput {
-            context_in: Some(unsafe {
-                std::slice::from_raw_parts_mut(
-                    &mut args as *mut _ as *mut u8,
-                    std::mem::size_of_val(&args),
-                )
-            }),
-            ..Default::default()
-        };
-
-        let output = skel.progs.arena_init.test_run(input)?;
-        if output.return_value != 0 {
-            bail!(
-                "Could not initialize arenas, p2dq_setup returned {}",
-                output.return_value as i32
-            );
-        }
-
-        Ok(())
-    }
-
-    fn setup_topology_node(&self, skel: &mut BpfSkel, mask: &[u64]) -> Result<()> {
-        let mut args = types::arena_alloc_mask_args {
-            bitmap: 0 as c_ulong,
-        };
-
-        let input = ProgramInput {
-            context_in: Some(unsafe {
-                std::slice::from_raw_parts_mut(
-                    &mut args as *mut _ as *mut u8,
-                    std::mem::size_of_val(&args),
-                )
-            }),
-            ..Default::default()
-        };
-
-        let output = skel.progs.arena_alloc_mask.test_run(input)?;
-        if output.return_value != 0 {
-            bail!(
-                "Could not initialize arenas, setup_topology_node returned {}",
-                output.return_value as i32
-            );
-        }
-
-        let ptr = unsafe { std::mem::transmute::<u64, &mut [u64; 10]>(args.bitmap) };
-
-        let (valid_mask, _) = ptr.split_at_mut(mask.len());
-        valid_mask.clone_from_slice(mask);
-
-        let mut args = types::arena_topology_node_init_args {
-            bitmap: args.bitmap as c_ulong,
-            data_size: 0 as c_ulong,
-            id: 0 as c_ulong,
-        };
-
-        let input = ProgramInput {
-            context_in: Some(unsafe {
-                std::slice::from_raw_parts_mut(
-                    &mut args as *mut _ as *mut u8,
-                    std::mem::size_of_val(&args),
-                )
-            }),
-            ..Default::default()
-        };
-
-        let output = skel.progs.arena_topology_node_init.test_run(input)?;
-        if output.return_value != 0 {
-            bail!(
-                "p2dq_topology_node_init returned {}",
-                output.return_value as i32
-            );
-        }
-
-        Ok(())
-    }
-
-    fn setup_topology(&self, skel: &mut BpfSkel) -> Result<()> {
-        let topo = Topology::new().expect("Failed to build host topology");
-
-        self.setup_topology_node(skel, topo.span.as_raw_slice())?;
-
-        for (_, node) in topo.nodes {
-            self.setup_topology_node(skel, node.span.as_raw_slice())?;
-        }
-
-        for (_, llc) in topo.all_llcs {
-            self.setup_topology_node(
-                skel,
-                Arc::<Llc>::into_inner(llc)
-                    .expect("missing llc")
-                    .span
-                    .as_raw_slice(),
-            )?;
-        }
-
-        for (_, core) in topo.all_cores {
-            self.setup_topology_node(
-                skel,
-                Arc::<Core>::into_inner(core)
-                    .expect("missing core")
-                    .span
-                    .as_raw_slice(),
-            )?;
-        }
-        for (_, cpu) in topo.all_cpus {
-            let mut mask = [0; 9];
-            mask[cpu.id.checked_shr(64).unwrap_or(0)] |= 1 << (cpu.id % 64);
-            self.setup_topology_node(skel, &mask)?;
-        }
-
-        Ok(())
-    }
-
     fn attach_kprobes(&self, skel: &mut BpfSkel) -> Result<Vec<Link>> {
         let Some(kd) = &self.kprobe_random_delays else {
             return Ok(vec![]);
@@ -425,9 +299,21 @@ impl Builder<'_> {
         skel_builder.obj_builder.debug(self.verbose > 1);
         init_libbpf_logging(None);
 
+        let topo = if self.p2dq_opts.virt_llc_enabled {
+            Topology::with_args(&self.p2dq_opts.topo)?
+        } else {
+            Topology::new()?
+        };
         let open_opts = LibbpfOpts::default().into_bpf_open_opts();
         let mut open_skel = scx_ops_open!(skel_builder, open_object, chaos, open_opts)?;
-        scx_p2dq::init_open_skel!(&mut open_skel, self.p2dq_opts, self.verbose)?;
+        let hw_profile = scx_p2dq::HardwareProfile::detect();
+        scx_p2dq::init_open_skel!(
+            &mut open_skel,
+            &topo,
+            self.p2dq_opts,
+            self.verbose,
+            &hw_profile
+        )?;
 
         let rodata = open_skel.maps.rodata_data.as_mut().unwrap();
 
@@ -526,10 +412,11 @@ impl Builder<'_> {
         };
 
         let mut skel = scx_ops_load!(open_skel, chaos, uei)?;
-        scx_p2dq::init_skel!(&mut skel);
+        scx_p2dq::init_skel!(&mut skel, topo);
 
-        self.setup_arenas(&mut skel)?;
-        self.setup_topology(&mut skel)?;
+        let task_size = std::mem::size_of::<types::task_p2dq>();
+        let arenalib = ArenaLib::init(skel.object_mut(), task_size, *NR_CPU_IDS)?;
+        arenalib.setup()?;
 
         let out = unsafe {
             // SAFETY: initialising field by field. open_object is already "initialised" (it's
