@@ -4,10 +4,15 @@
 // GNU General Public License version 2.
 
 fn main() {
-    scx_utils::BpfBuilder::new()
+    scx_cargo::BpfBuilder::new()
         .unwrap()
         .enable_intf("src/bpf/intf.h", "bpf_intf.rs")
         .enable_skel("src/bpf/main.bpf.c", "bpf")
-        .build()
+        .add_source("src/bpf/introspec.bpf.c")
+        .add_source("src/bpf/lock.bpf.c")
+        .add_source("src/bpf/power.bpf.c")
+        .add_source("src/bpf/preempt.bpf.c")
+        .add_source("src/bpf/sys_stat.bpf.c")
+        .compile_link_gen()
         .unwrap();
 }
