@@ -8,6 +8,8 @@
  * To be included to the main.bpf.c
  */
 
+extern const volatile u8	mig_delta_pct;
+
 u64 __attribute__ ((noinline)) calc_mig_delta(u64 avg_sc_load, int nz_qlen)
 {
 	/*
@@ -66,7 +68,13 @@ int plan_x_cpdom_migration(void)
 			continue;
 		}
 
-		util = (cpdomc->cur_util_sum << LAVD_SHIFT) / cpdomc->nr_active_cpus;
+		/*
+		 * Use avg_util_sum when mig_delta_pct is set, otherwise use cur_util_sum.
+		 */
+		if (mig_delta_pct > 0)
+			util = (cpdomc->avg_util_sum << LAVD_SHIFT) / cpdomc->nr_active_cpus;
+		else
+			util = (cpdomc->cur_util_sum << LAVD_SHIFT) / cpdomc->nr_active_cpus;
 		qlen = cpdomc->nr_queued_task;
 		sc_qlen = (qlen << (LAVD_SHIFT * 3)) / cpdomc->cap_sum_active_cpus;
 		cpdomc->sc_load = util + sc_qlen;
@@ -85,8 +93,15 @@ int plan_x_cpdom_migration(void)
 	/*
 	 * Determine the criteria for stealer and stealee domains.
 	 * The more the system is loaded, the tighter criteria will be chosen.
+	 * When mig_delta_pct is set (non-zero), use it as a fixed percentage
+	 * instead of the dynamic calculation.
 	 */
-	x_mig_delta = calc_mig_delta(avg_sc_load, nz_qlen);
+	if (mig_delta_pct > 0) {
+		u64 mig_delta_factor = (mig_delta_pct << LAVD_SHIFT) / 100;
+		x_mig_delta = avg_sc_load * mig_delta_factor / LAVD_SCALE;
+	} else {
+		x_mig_delta = calc_mig_delta(avg_sc_load, nz_qlen);
+	}
 	stealer_threshold = avg_sc_load - x_mig_delta;
 	stealee_threshold = avg_sc_load + x_mig_delta;
 
@@ -351,7 +366,7 @@ static bool consume_task(u64 cpu_dsq_id, u64 cpdom_dsq_id)
 {
 	struct cpdom_ctx *cpdomc;
 	struct task_struct *p;
-	u64 vtime = U64_MAX, dsq_id = cpu_dsq_id;
+	u64 vtime = U64_MAX;
 
 	cpdomc = MEMBER_VPTR(cpdom_ctxs, [dsq_to_cpdom(cpdom_dsq_id)]);
 	if (!cpdomc) {
@@ -367,32 +382,47 @@ static bool consume_task(u64 cpu_dsq_id, u64 cpdom_dsq_id)
 	    try_to_steal_task(cpdomc))
 		goto x_domain_migration_out;
 
-	if (per_cpu_dsq) {
-		bpf_for_each(scx_dsq, p, cpu_dsq_id, 0) {
-			vtime = p->scx.dsq_vtime;
-			break;
-		}
-
-		bpf_for_each(scx_dsq, p, cpdom_dsq_id, 0) {
-			if (p->scx.dsq_vtime < vtime)
-				dsq_id = cpdom_dsq_id;
-			break;
-		}
-	} else {
-		dsq_id = cpdom_dsq_id;
-	}
-
 	/*
-	 * Try to consume a task from selected DSQ.
+	 * When per_cpu_dsq or pinned_slice_ns is enabled, compare vtimes
+	 * across cpu_dsq and cpdom_dsq to select the task with the lowest vtime.
 	 */
-	if (consume_dsq(cpdomc, dsq_id))
-		return true;
+	if (per_cpu_dsq || pinned_slice_ns) {
+		u64 dsq_id = cpu_dsq_id;
+		u64 backup_dsq_id = cpdom_dsq_id;
+
+		p = __COMPAT_scx_bpf_dsq_peek(cpu_dsq_id);
+		if (p)
+			vtime = p->scx.dsq_vtime;
+
+		p = __COMPAT_scx_bpf_dsq_peek(cpu_dsq_id);
+		if (p && p->scx.dsq_vtime < vtime) {
+			dsq_id = cpdom_dsq_id;
+			backup_dsq_id = cpu_dsq_id;
+		}
+
+		/*
+		 * There is a scenario where the task on the Cpdom DSQ has a
+		 * lower vtime, but this CPU fails to win the race and causes
+		 * the pinned task to stall and wait on the Per-CPU DSQ for the
+		 * next scheduling round. Always try consuming from the other DSQ
+		 * to prevent this scenario.
+		 */
+		if (consume_dsq(cpdomc, dsq_id))
+			return true;
+		if (consume_dsq(cpdomc, backup_dsq_id))
+			return true;
+	} else {
+		if (consume_dsq(cpdomc, cpdom_dsq_id))
+			return true;
+	}
 
 	/*
 	 * If there is no task in the assssociated DSQ, traverse neighbor
 	 * compute domains in distance order -- task stealing.
+	 * Skip force stealing when mig_delta_pct is set (> 0) to rely
+	 * only on the is_stealer/is_stealee thresholds.
 	 */
-	if (nr_cpdoms > 1 && force_to_steal_task(cpdomc))
+	if (nr_cpdoms > 1 && mig_delta_pct == 0 && force_to_steal_task(cpdomc))
 		goto x_domain_migration_out;
 
 	return false;
