@@ -208,6 +208,18 @@ static u64		cur_svc_time;
 const volatile u64	slice_min_ns = LAVD_SLICE_MIN_NS_DFL;
 const volatile u64	slice_max_ns = LAVD_SLICE_MAX_NS_DFL;
 
+/*
+ * Migration delta threshold percentage (0-100)
+ */
+const volatile u8	mig_delta_pct = 0;
+
+/*
+ * Slice time for all tasks when pinned tasks are running on the CPU.
+ * When this is set (non-zero), pinned tasks always use per-CPU DSQs and
+ * the dispatch logic compares vtimes across DSQs.
+ */
+const volatile u64	pinned_slice_ns = 0;
+
 static volatile u64	nr_cpus_big;
 
 /*
@@ -261,6 +273,17 @@ static u64 calc_time_slice(struct task_ctx *taskc, struct cpu_ctx *cpuc)
 	if (!taskc || !cpuc)
 		return LAVD_SLICE_MAX_NS_DFL;
 
+	/*
+	 * If pinned_slice_ns is enabled and there are pinned tasks waiting
+	 * to run on this CPU, unconditionally reduce the time slice for
+	 * all tasks to ensure pinned tasks can run promptly.
+	 */
+	if (pinned_slice_ns && cpuc->nr_pinned_tasks) {
+		taskc->slice = pinned_slice_ns;
+		reset_task_flag(taskc, LAVD_FLAG_SLICE_BOOST);
+		return taskc->slice;
+	}
+
 	/*
 	 * If the task's avg_runtime is greater than the regular time slice
 	 * (i.e., taskc->avg_runtime > sys_stat.slice), that means the task
@@ -510,7 +533,7 @@ s32 BPF_STRUCT_OPS(lavd_select_cpu, struct task_struct *p, s32 prev_cpu,
 	struct task_ctx *taskc = get_task_ctx(p);
 	struct cpu_ctx *cpuc_cur = get_cpu_ctx();
 	struct cpu_ctx *cpuc;
-	u64 dsq_id;
+	u64 dsq_id, nr_queued = 0;
 	s32 cpu_id;
 	struct pick_ctx ictx = {
 		.p = p,
@@ -550,12 +573,12 @@ s32 BPF_STRUCT_OPS(lavd_select_cpu, struct task_struct *p, s32 prev_cpu,
 			goto out;
 		}
 
-		if (per_cpu_dsq)
-			dsq_id = cpu_to_dsq(cpu_id);
-		else
-			dsq_id = cpdom_to_dsq(cpuc->cpdom_id);
+		dsq_id = (per_cpu_dsq || pinned_slice_ns) ? cpu_to_dsq(cpu_id) : cpdom_to_dsq(cpuc->cpdom_id);
+		nr_queued = scx_bpf_dsq_nr_queued(dsq_id);
+		if (pinned_slice_ns)
+			nr_queued += scx_bpf_dsq_nr_queued(cpdom_to_dsq(cpuc->cpdom_id));
 
-		if (!scx_bpf_dsq_nr_queued(dsq_id)) {
+		if (!nr_queued) {
 			p->scx.dsq_vtime = calc_when_to_run(p, taskc);
 			p->scx.slice = LAVD_SLICE_MAX_NS_DFL;
 			scx_bpf_dsq_insert(p, SCX_DSQ_LOCAL, p->scx.slice, 0);
@@ -582,8 +605,8 @@ void BPF_STRUCT_OPS(lavd_enqueue, struct task_struct *p, u64 enq_flags)
 	struct task_ctx *taskc;
 	struct cpu_ctx *cpuc, *cpuc_cur;
 	s32 task_cpu, cpu = -ENOENT;
-	u64 cpdom_id;
-	bool is_idle = false;
+	u64 cpdom_id, dsq_id;
+	bool is_idle = false, can_direct;
 
 	taskc = get_task_ctx(p);
 	cpuc_cur = get_cpu_ctx();
@@ -650,15 +673,21 @@ void BPF_STRUCT_OPS(lavd_enqueue, struct task_struct *p, u64 enq_flags)
 	 * Enqueue the task to a DSQ. If it is safe to directly dispatch
 	 * to the local DSQ of the chosen CPU, do it. Otherwise, enqueue
 	 * to the chosen DSQ of the chosen domain.
+	 *
+	 * When pinned_slice_ns is enabled, pinned tasks always use per-CPU DSQ
+	 * to enable vtime comparison across DSQs during dispatch.
 	 */
-	if (can_direct_dispatch(cpu_to_dsq(cpu), cpu, is_idle)) {
+	dsq_id = (per_cpu_dsq || (pinned_slice_ns && is_pinned(p))) ? cpu_to_dsq(cpu) : cpdom_to_dsq(cpdom_id);
+
+	can_direct = can_direct_dispatch(dsq_id, cpu, is_idle);
+	if (can_direct && pinned_slice_ns && is_pinned(p))
+		can_direct &= can_direct_dispatch(cpdom_to_dsq(cpdom_id), cpu, is_idle);
+
+	if (can_direct) {
 		scx_bpf_dsq_insert(p, SCX_DSQ_LOCAL_ON | cpu, p->scx.slice,
 				   enq_flags);
-	} else if (per_cpu_dsq) {
-		scx_bpf_dsq_insert_vtime(p, cpu_to_dsq(cpu), p->scx.slice,
-					 p->scx.dsq_vtime, enq_flags);
 	} else {
-		scx_bpf_dsq_insert_vtime(p, cpdom_to_dsq(cpdom_id), p->scx.slice,
+		scx_bpf_dsq_insert_vtime(p, dsq_id, p->scx.slice,
 					 p->scx.dsq_vtime, enq_flags);
 	}
 
@@ -1005,12 +1034,6 @@ void BPF_STRUCT_OPS(lavd_running, struct task_struct *p)
 		return;
 	}
 
-	/*
-	 * Increase the number of pinned tasks waiting for execution.
-	 */
-	if (is_pinned(p))
-		__sync_fetch_and_sub(&cpuc->nr_pinned_tasks, 1);
-
 	/*
 	 * If the sched_ext core directly dispatched a task, calculating the
 	 * task's deadline and time slice was also skipped. In this case, we
@@ -1072,6 +1095,20 @@ void BPF_STRUCT_OPS(lavd_tick, struct task_struct *p)
 	now = scx_bpf_now();
 	account_task_runtime(p, taskc, cpuc, now);
 
+	/*
+	 * If pinned_slice_ns is enabled and there are pinned tasks waiting
+	 * to run on this CPU, unconditionally reduce the time slice for
+	 * all tasks to ensure pinned tasks can run promptly.
+	 */
+	if (pinned_slice_ns && cpuc->nr_pinned_tasks &&
+	    p->scx.slice > pinned_slice_ns) {
+		if (taskc->acc_runtime > pinned_slice_ns)
+			p->scx.slice = 0;
+		else
+			p->scx.slice = pinned_slice_ns;
+		return;
+	}
+
 	/*
 	 * If this task is slice-boosted and there is a pinned task that
 	 * must run on this, shrink its time slice to the regular one.
@@ -1131,6 +1168,13 @@ void BPF_STRUCT_OPS(lavd_quiescent, struct task_struct *p, u64 deq_flags)
 		taskc->wait_freq = calc_avg_freq(taskc->wait_freq, interval);
 		taskc->last_quiescent_clk = now;
 	}
+
+	/*
+	 * Decrease the number of pinned tasks waiting for execution.
+	 */
+	if (is_pinned(p))
+		__sync_fetch_and_sub(&cpuc->nr_pinned_tasks, 1);
+
 }
 
 static void cpu_ctx_init_online(struct cpu_ctx *cpuc, u32 cpu_id, u64 now)
@@ -1362,7 +1406,7 @@ void BPF_STRUCT_OPS(lavd_cpu_release, s32 cpu,
 
 	/*
 	 * Keep track of when the higher-priority scheduler class takes
-	 * the CPUto calculate capacity-invariant and frequency-invariant
+	 * the CPU to calculate capacity-invariant and frequency-invariant
 	 * CPU utilization.
 	 */
 	cpuc->cpu_release_clk = scx_bpf_now();
@@ -1773,8 +1817,10 @@ s32 BPF_STRUCT_OPS_SLEEPABLE(lavd_init)
 
 	/*
 	 * Initialize per-CPU DSQs.
+	 * Per-CPU DSQs are created when per_cpu_dsq is enabled OR when
+	 * pinned_slice_ns is enabled (for pinned task handling).
 	 */
-	if (per_cpu_dsq) {
+	if (per_cpu_dsq || pinned_slice_ns) {
 		err = init_per_cpu_dsqs();
 		if (err)
 			return err;
