@@ -251,7 +251,6 @@ int do_core_compaction(void)
 	struct cpdom_ctx *cpdomc;
 	int nr_active, cpu, i;
 	u32 sum_capacity = 0, big_capacity = 0, nr_active_cpdoms = 0;
-	bool need_kick;
 	u64 cpdom_id;
 
 	bpf_rcu_read_lock();
@@ -296,11 +295,9 @@ int do_core_compaction(void)
 		/*
 		 * Assign an online cpu to active and overflow cpumasks
 		 */
-		need_kick = false;
 		if (i < nr_active) {
 			bpf_cpumask_set_cpu(cpu, active);
 			bpf_cpumask_clear_cpu(cpu, ovrflw);
-			need_kick = true;
 
 			/*
 			 * Accumulate the capacity of active CPUs and
@@ -312,12 +309,6 @@ int do_core_compaction(void)
 				cpdomc->nr_acpus_temp++;
 			}
 
-			/*
-			 * Calculate big capacity ratio among active cores.
-			 */
-			sum_capacity += cpuc->capacity;
-			if (cpuc->big_core)
-				big_capacity += cpuc->capacity;
 		} else {
 			bpf_cpumask_clear_cpu(cpu, active);
 
@@ -328,7 +319,6 @@ int do_core_compaction(void)
 				 * add this CPU{ to the overflow set.
 				 */
 				bpf_cpumask_set_cpu(cpu, ovrflw);
-				need_kick = true;
 			} else if ((bpf_get_prandom_u32() %
 				    LAVD_CC_CPU_PIN_INTERVAL_DIV)) {
 				/*
@@ -346,16 +336,23 @@ int do_core_compaction(void)
 				 * approximately for LAVD_CC_CPU_PIN_INTERVAL.
 				 */
 				bpf_cpumask_clear_cpu(cpu, ovrflw);
-			} else if (bpf_cpumask_test_cpu(cpu, cast_mask(ovrflw))) {
-				need_kick = true;
+				continue;
+			} else if (!bpf_cpumask_test_cpu(cpu, cast_mask(ovrflw))) {
+				continue;
 			}
 		}
 
 		/*
 		 * When the CPU is in either an active or overflow set, kick it.
 		 */
-		if (need_kick)
-			scx_bpf_kick_cpu(cpu, SCX_KICK_IDLE);
+		scx_bpf_kick_cpu(cpu, SCX_KICK_IDLE);
+
+		/*
+		 * Calculate big capacity ratio if a CPU is on.
+		 */
+		sum_capacity += cpuc->capacity;
+		if (cpuc->big_core)
+			big_capacity += cpuc->capacity;
 	}
 
 	cur_big_core_scale = (big_capacity << LAVD_SHIFT) / sum_capacity;
@@ -557,7 +554,7 @@ int update_thr_perf_cri(void)
 		 *   |     little_core_scale
 		 *   0
 		 */
-		delta = sys_stat.avg_perf_cri - sys_stat.min_perf_cri;
+		delta = (sys_stat.avg_perf_cri - sys_stat.min_perf_cri) >> 1;
 		diff = (delta * little_core_scale) >> LAVD_SHIFT;
 		thr = diff + sys_stat.min_perf_cri;
 	}
@@ -574,10 +571,14 @@ int update_thr_perf_cri(void)
 		 *   |                     |           1024
 		 *   |                     little_core_scale
 		 *   0
+		 *
+		 *  Note that half of the little core capacity is taken by the
+		 *  [min_perf_cri, avg_perf_cri] range, so only another half
+		 *  can serve for the [avg_perf_cri, max_perf_cri range.
 		 */
-		delta = sys_stat.max_perf_cri - sys_stat.avg_perf_cri;
-		diff = (delta * cur_big_core_scale) >> LAVD_SHIFT;
-		thr = sys_stat.max_perf_cri - diff;
+		delta = (sys_stat.max_perf_cri - sys_stat.avg_perf_cri) >> 1;
+		diff = (delta * (little_core_scale >> 1)) >> LAVD_SHIFT;
+		thr = diff + sys_stat.avg_perf_cri;
 	}
 
 	sys_stat.thr_perf_cri = thr;
