@@ -55,6 +55,7 @@ use scx_utils::try_set_rlimit_infinity;
 use scx_utils::uei_exited;
 use scx_utils::uei_report;
 use scx_utils::EnergyModel;
+use scx_utils::TopologyArgs;
 use scx_utils::UserExitInfo;
 use scx_utils::NR_CPU_IDS;
 use stats::SchedSample;
@@ -118,6 +119,24 @@ struct Opts {
     #[clap(long = "slice-min-us", default_value = "500")]
     slice_min_us: u64,
 
+    /// Migration delta threshold percentage (0-100). When set to a non-zero value,
+    /// uses average utilization for threshold calculation instead of current
+    /// utilization, and the threshold is calculated as: avg_load * (mig-delta-pct / 100).
+    /// Additionally, disables force task stealing in the consume path, relying only
+    /// on the is_stealer/is_stealee thresholds for more predictable load balancing.
+    /// Default is 0 (disabled, uses dynamic threshold based on load with both
+    /// probabilistic and force task stealing enabled). This is an experimental feature.
+    #[clap(long = "mig-delta-pct", default_value = "0", value_parser=Opts::mig_delta_pct_range)]
+    mig_delta_pct: u8,
+
+    /// Slice duration in microseconds to use for all tasks when pinned tasks
+    /// are running on a CPU. Must be between slice-min-us and slice-max-us.
+    /// When this option is enabled, pinned tasks are always enqueued to per-CPU DSQs
+    /// and the dispatch logic compares vtimes across all DSQs to select the lowest
+    /// vtime task. This helps improve responsiveness for pinned tasks.
+    #[clap(long = "pinned-slice-us")]
+    pinned_slice_us: Option<u64>,
+
     /// Limit the ratio of preemption to the roughly top P% of latency-critical
     /// tasks. When N is given as an argument, P is 0.5^N * 100. The default
     /// value is 6, which limits the preemption for the top 1.56% of
@@ -205,6 +224,10 @@ struct Opts {
 
     #[clap(flatten, next_help_heading = "Libbpf Options")]
     pub libbpf: LibbpfOpts,
+
+    /// Topology configuration options
+    #[clap(flatten)]
+    topology: Option<TopologyArgs>,
 }
 
 impl Opts {
@@ -300,12 +323,30 @@ impl Opts {
             info!("Energy model won't be used for CPU preference order.");
         }
 
+        if let Some(pinned_slice) = self.pinned_slice_us {
+            if pinned_slice < self.slice_min_us || pinned_slice > self.slice_max_us {
+                info!(
+                    "pinned-slice-us ({}) must be between slice-min-us ({}) and slice-max-us ({})",
+                    pinned_slice, self.slice_min_us, self.slice_max_us
+                );
+                return None;
+            }
+            info!(
+                "Pinned task slice mode is enabled ({} us). Pinned tasks will use per-CPU DSQs.",
+                pinned_slice
+            );
+        }
+
         Some(self)
     }
 
     fn preempt_shift_range(s: &str) -> Result<u8, String> {
         number_range(s, 0, 10)
     }
+
+    fn mig_delta_pct_range(s: &str) -> Result<u8, String> {
+        number_range(s, 0, 100)
+    }
 }
 
 unsafe impl Plain for msg_task_ctx {}
@@ -364,8 +405,8 @@ impl<'a> Scheduler<'a> {
             }
         }
 
-        // Initialize CPU topology
-        let order = CpuOrder::new().unwrap();
+        // Initialize CPU topology with CLI arguments
+        let order = CpuOrder::new(opts.topology.as_ref()).unwrap();
         Self::init_cpus(&mut skel, &order);
         Self::init_cpdoms(&mut skel, &order);
 
@@ -414,6 +455,11 @@ impl<'a> Scheduler<'a> {
             ("futex_unlock_pi", &skel.progs.fexit_futex_unlock_pi),
         ];
 
+        if compat::tracer_available("function")? == false {
+            info!("Ftrace is not enabled in the kernel.");
+            return Ok(false);
+        }
+
         compat::cond_kprobes_enable(ftraces)
     }
 
@@ -536,7 +582,9 @@ impl<'a> Scheduler<'a> {
         rodata.verbose = opts.verbose;
         rodata.slice_max_ns = opts.slice_max_us * 1000;
         rodata.slice_min_ns = opts.slice_min_us * 1000;
+        rodata.pinned_slice_ns = opts.pinned_slice_us.map(|v| v * 1000).unwrap_or(0);
         rodata.preempt_shift = opts.preempt_shift;
+        rodata.mig_delta_pct = opts.mig_delta_pct;
         rodata.no_use_em = opts.no_use_em as u8;
         rodata.no_wake_sync = opts.no_wake_sync;
         rodata.no_slice_boost = opts.no_slice_boost;
