@@ -36,6 +36,7 @@ enum consts {
 	/* 64 chars for user-provided name, 64 for possible template suffix. */
 	MAX_LAYER_NAME		= 128,
 	MAX_LAYERS		= 16,
+	MAX_CGROUP_REGEXES	= 16,
 	MAX_LAYER_WEIGHT	= 10000,
 	MIN_LAYER_WEIGHT	= 1,
 	DEFAULT_LAYER_WEIGHT	= 100,
@@ -67,6 +68,13 @@ enum consts {
 	SCXCMD_PREFIX		= 0x5C10,
 	SCXCMD_COMLEN		= 13,
 	MAX_GPU_PIDS 		= 100000,
+
+};
+
+enum layer_membership {
+	MEMBER_NOEXPIRE		= (u64)-1,
+	MEMBER_EXPIRED		= (u64)-2,
+	MEMBER_CANTMATCH	= (u64)-3,
 };
 
 static inline void ___consts_sanity_check___(void) {
@@ -75,6 +83,8 @@ static inline void ___consts_sanity_check___(void) {
 	_Static_assert(MAX_LLCS <= (1 << DSQ_ID_LAYER_SHIFT), "MAX_LLCS too high");
 	_Static_assert(MAX_LAYERS <= (DSQ_ID_LAYER_MASK >> DSQ_ID_LAYER_SHIFT) + 1,
 		       "MAX_LAYERS too high");
+	/* cgroup regex matching uses u64 as match bitmap */
+	_Static_assert(MAX_CGROUP_REGEXES <= 64, "MAX_CGROUP_REGEXES too high for u64 bitmap");
 }
 
 enum layer_kind {
@@ -182,6 +192,7 @@ struct cpu_ctx {
 	bool			is_protected;
 
 	u64			layer_usages[MAX_LAYERS][NR_LAYER_USAGES];
+	u64			layer_membw_agg[MAX_LAYERS][NR_LAYER_USAGES];
 	u64			gstats[NR_GSTATS];
 	u64			lstats[MAX_LAYERS][NR_LSTATS];
 	u64			ran_current_for;
@@ -193,7 +204,6 @@ struct cpu_ctx {
 	u64			lo_fb_dsq_id;
 	bool			in_open_layers;
 	u32			layer_id;
-	u32			task_layer_id;
 	u32			llc_id;
 	u32			node_id;
 	u32			perf;
@@ -260,7 +270,10 @@ enum layer_match_kind {
 	MATCH_AVG_RUNTIME,
 	MATCH_CGROUP_SUFFIX,
 	MATCH_CGROUP_CONTAINS,
+	MATCH_CGROUP_REGEX,
 	MATCH_HINT_EQUALS,
+	MATCH_SYSTEM_CPU_UTIL_BELOW,
+	MATCH_DSQ_INSERT_BELOW,
 
 	NR_LAYER_MATCH_KINDS,
 };
@@ -270,6 +283,7 @@ struct layer_match {
 	char		cgroup_prefix[MAX_PATH];
 	char		cgroup_suffix[MAX_PATH];
 	char		cgroup_substr[MAX_PATH];
+	u32		cgroup_regex_id;
 	char		comm_prefix[MAX_COMM];
 	char		pcomm_prefix[MAX_COMM];
 	int		nice;
@@ -287,6 +301,8 @@ struct layer_match {
 	u64		min_avg_runtime_us;
 	u64		max_avg_runtime_us;
 	u64		hint;
+	u64		system_cpu_util_below;	/* ratio * 10000 */
+	u64		dsq_insert_below;	/* ratio * 10000 */
 };
 
 struct layer_match_ands {
@@ -363,6 +379,7 @@ struct layer {
 	bool			is_protected;
 	bool			periodically_refresh;
 	u8			cpuset[MAX_CPUS_U8];
+	u64			member_expire_ms;
 };
 
 struct scx_cmd {
@@ -371,4 +388,10 @@ struct scx_cmd {
 	u8			cmd[SCXCMD_COMLEN];
 } __attribute__((packed));
 
+struct hint_layer_info {
+	u32			layer_id;
+	u64			system_cpu_util_below;	/* ratio * 10000, u64::MAX = disabled */
+	u64			dsq_insert_below;	/* ratio * 10000, u64::MAX = disabled */
+};
+
 #endif /* __INTF_H */
