@@ -24,6 +24,7 @@ use serde::Serialize;
 use crate::bpf_intf;
 use crate::BpfStats;
 use crate::Layer;
+use crate::LayerKind;
 use crate::Stats;
 use crate::LAYER_USAGE_OPEN;
 use crate::LAYER_USAGE_PROTECTED;
@@ -209,6 +210,10 @@ pub struct LayerStats {
     pub llc_fracs: Vec<f64>,
     #[stat(desc = "Per-LLC average latency")]
     pub llc_lats: Vec<f64>,
+    #[stat(desc = "Layer memory bandwidth as a % of total allowed (0 for \"no limit\"")]
+    pub membw_pct: f64,
+    #[stat(desc = "DSQ insertion ratio EWMA (10s window)")]
+    pub dsq_insert_ewma: f64,
 }
 
 impl LayerStats {
@@ -239,6 +244,23 @@ impl LayerStats {
             .take(LAYER_USAGE_SUM_UPTO + 1)
             .sum::<f64>();
 
+        let membw_frac = match &layer.kind {
+            // Open layer's can't have a memory BW limit.
+            LayerKind::Open { .. } => 0.0,
+            LayerKind::Confined { membw_gb, .. } | LayerKind::Grouped { membw_gb, .. } => {
+                // Check if we have set a memory BW limit.
+                if let Some(membw_limit_gb) = membw_gb {
+                    stats.layer_membws[lidx]
+                        .iter()
+                        .take(LAYER_USAGE_SUM_UPTO + 1)
+                        .sum::<f64>()
+                        / ((*membw_limit_gb * (1024_u64.pow(3) as f64)) as f64)
+                } else {
+                    0.0
+                }
+            }
+        };
+
         Self {
             index: lidx,
             util: util_sum * 100.0,
@@ -308,6 +330,8 @@ impl LayerStats {
                 .iter()
                 .map(|lstats| lstats[LLC_LSTAT_LAT] as f64 / 1_000_000_000.0)
                 .collect(),
+            membw_pct: membw_frac * 100.0,
+            dsq_insert_ewma: stats.layer_dsq_insert_ewma[lidx] * 100.0,
         }
     }
 
@@ -327,7 +351,7 @@ impl LayerStats {
 
         writeln!(
             w,
-            "  {:<width$}  tot={:7} local_sel/enq={}/{} enq_dsq={} wake/exp/reenq={}/{}/{}",
+            "  {:<width$}  tot={:7} local_sel/enq={}/{} enq_dsq={} wake/exp/reenq={}/{}/{} dsq_ewma={}",
             "",
             self.total,
             fmt_pct(self.sel_local),
@@ -336,6 +360,7 @@ impl LayerStats {
             fmt_pct(self.enq_wakeup),
             fmt_pct(self.enq_expire),
             fmt_pct(self.enq_reenq),
+            fmt_pct(self.dsq_insert_ewma),
             width = header_width,
         )?;
 
@@ -518,6 +543,8 @@ pub struct SysStats {
     pub gpu_tasks_affinitized: u64,
     #[stat(desc = "Time (in ms) of last affinitization run.")]
     pub gpu_task_affinitization_ms: u64,
+    #[stat(desc = "System CPU utilization EWMA (10s window)")]
+    pub system_cpu_util_ewma: f64,
 }
 
 impl SysStats {
@@ -577,6 +604,7 @@ impl SysStats {
             layers: BTreeMap::new(),
             gpu_tasks_affinitized: stats.gpu_tasks_affinitized,
             gpu_task_affinitization_ms: stats.gpu_task_affinitization_ms,
+            system_cpu_util_ewma: stats.system_cpu_util_ewma * 100.0,
         })
     }
 
@@ -595,7 +623,7 @@ impl SysStats {
 
         writeln!(
             w,
-            "busy={:5.1} util/hi/lo={:7.1}/{}/{} fallback_cpu/util={:3}/{:4.1} proc={:?}ms",
+            "busy={:5.1} util/hi/lo={:7.1}/{}/{} fallback_cpu/util={:3}/{:4.1} proc={:?}ms sys_util_ewma={:5.1}",
             self.busy,
             self.util,
             fmt_pct(self.hi_fb_util),
@@ -603,6 +631,7 @@ impl SysStats {
             self.fallback_cpu,
             self.fallback_cpu_util,
             self.proc_ms,
+            self.system_cpu_util_ewma,
         )?;
 
         writeln!(
@@ -709,7 +738,7 @@ pub fn server_data() -> StatsServerData<StatsReq, StatsRes> {
 
 pub fn monitor(intv: Duration, shutdown: Arc<AtomicBool>) -> Result<()> {
     scx_utils::monitor_stats::<SysStats>(
-        &vec![],
+        &[],
         intv,
         || shutdown.load(Ordering::Relaxed),
         |sst| {
