@@ -28,12 +28,15 @@ char _license[] SEC("license") = "GPL";
 /*
  * Variables populated by userspace
  */
-const volatile u32 nr_possible_cpus = 1;
-const volatile bool smt_enabled = true;
+const volatile u32	     nr_possible_cpus = 1;
+const volatile bool	     smt_enabled      = true;
 const volatile unsigned char all_cpus[MAX_CPUS_U8];
 
-const volatile u64 slice_ns;
-const volatile u64 root_cgid = 1;
+const volatile u64	     slice_ns;
+const volatile u64	     root_cgid			     = 1;
+const volatile bool	     debug_events_enabled	     = false;
+const volatile bool	     exiting_task_workaround_enabled = true;
+const volatile bool	     split_vtime_updates	     = false;
 
 /*
  * CPU assignment changes aren't fully in effect until a subsequent tick()
@@ -43,6 +46,29 @@ const volatile u64 root_cgid = 1;
 u32 configuration_seq;
 u32 applied_configuration_seq;
 
+/*
+ * Debug events circular buffer
+ */
+u32 debug_event_pos;
+
+struct {
+	__uint(type, BPF_MAP_TYPE_ARRAY);
+	__uint(max_entries, DEBUG_EVENTS_BUF_SIZE);
+	__type(key, u32);
+	__type(value, struct debug_event);
+} debug_events SEC(".maps");
+
+struct update_timer {
+	struct bpf_timer timer;
+};
+
+struct {
+	__uint(type, BPF_MAP_TYPE_ARRAY);
+	__uint(max_entries, 1);
+	__type(key, u32);
+	__type(value, struct update_timer);
+} update_timer SEC(".maps") __weak;
+
 private(all_cpumask) struct bpf_cpumask __kptr *all_cpumask;
 private(root_cgrp) struct cgroup __kptr *root_cgrp;
 
@@ -68,7 +94,7 @@ static inline u32 dsq_to_cpu(u32 dsq)
 }
 
 static inline struct cgroup *lookup_cgrp_ancestor(struct cgroup *cgrp,
-						  u32 ancestor)
+						  u32		 ancestor)
 {
 	struct cgroup *cg;
 
@@ -86,7 +112,7 @@ struct {
 	__uint(map_flags, BPF_F_NO_PREALLOC);
 	__type(key, int);
 	__type(value, struct cgrp_ctx);
-} cgrp_ctxs SEC(".maps");
+} cgrp_ctxs		       SEC(".maps");
 
 static inline struct cgrp_ctx *lookup_cgrp_ctx_fallible(struct cgroup *cgrp)
 {
@@ -146,7 +172,7 @@ struct {
 	__uint(map_flags, BPF_F_NO_PREALLOC);
 	__type(key, int);
 	__type(value, struct task_ctx);
-} task_ctxs SEC(".maps");
+} task_ctxs		       SEC(".maps");
 
 static inline struct task_ctx *lookup_task_ctx(struct task_struct *p)
 {
@@ -165,12 +191,12 @@ struct {
 	__type(key, u32);
 	__type(value, struct cpu_ctx);
 	__uint(max_entries, 1);
-} cpu_ctxs SEC(".maps");
+} cpu_ctxs		      SEC(".maps");
 
 static inline struct cpu_ctx *lookup_cpu_ctx(int cpu)
 {
 	struct cpu_ctx *cctx;
-	u32 zero = 0;
+	u32		zero = 0;
 
 	if (cpu < 0)
 		cctx = bpf_map_lookup_elem(&cpu_ctxs, &zero);
@@ -185,7 +211,7 @@ static inline struct cpu_ctx *lookup_cpu_ctx(int cpu)
 	return cctx;
 }
 
-struct cell cells[MAX_CELLS];
+struct cell		   cells[MAX_CELLS];
 
 static inline struct cell *lookup_cell(int idx)
 {
@@ -212,8 +238,10 @@ static inline int allocate_cell()
 		if (!(c = lookup_cell(cell_idx)))
 			return -1;
 
-		if (__sync_bool_compare_and_swap(&c->in_use, 0, 1))
+		if (__sync_bool_compare_and_swap(&c->in_use, 0, 1)) {
+			WRITE_ONCE(c->vtime_now, 0);
 			return cell_idx;
+		}
 	}
 	scx_bpf_error("No available cells to allocate");
 	return -1;
@@ -235,6 +263,70 @@ static inline int free_cell(int cell_idx)
 	return 0;
 }
 
+/*
+ * Record debug events to the circular buffer
+ */
+static inline void record_cgroup_init(u64 cgid)
+{
+	struct debug_event *event;
+	u32		    pos, idx;
+
+	if (likely(!debug_events_enabled))
+		return;
+
+	pos   = __sync_fetch_and_add(&debug_event_pos, 1);
+	idx   = pos % DEBUG_EVENTS_BUF_SIZE;
+
+	event = bpf_map_lookup_elem(&debug_events, &idx);
+	if (unlikely(!event))
+		return;
+
+	event->timestamp	= scx_bpf_now();
+	event->event_type	= DEBUG_EVENT_CGROUP_INIT;
+	event->cgroup_init.cgid = cgid;
+}
+
+static inline void record_init_task(u64 cgid, u32 pid)
+{
+	struct debug_event *event;
+	u32		    pos, idx;
+
+	if (likely(!debug_events_enabled))
+		return;
+
+	pos   = __sync_fetch_and_add(&debug_event_pos, 1);
+	idx   = pos % DEBUG_EVENTS_BUF_SIZE;
+
+	event = bpf_map_lookup_elem(&debug_events, &idx);
+	if (unlikely(!event))
+		return;
+
+	event->timestamp      = scx_bpf_now();
+	event->event_type     = DEBUG_EVENT_INIT_TASK;
+	event->init_task.cgid = cgid;
+	event->init_task.pid  = pid;
+}
+
+static inline void record_cgroup_exit(u64 cgid)
+{
+	struct debug_event *event;
+	u32		    pos, idx;
+
+	if (likely(!debug_events_enabled))
+		return;
+
+	pos   = __sync_fetch_and_add(&debug_event_pos, 1);
+	idx   = pos % DEBUG_EVENTS_BUF_SIZE;
+
+	event = bpf_map_lookup_elem(&debug_events, &idx);
+	if (unlikely(!event))
+		return;
+
+	event->timestamp	= scx_bpf_now();
+	event->event_type	= DEBUG_EVENT_CGROUP_EXIT;
+	event->cgroup_exit.cgid = cgid;
+}
+
 /*
  * Store the cpumask for each cell (owned by BPF logic). We need this in an
  * explicit map to allow for these to be kptrs.
@@ -254,7 +346,7 @@ struct {
 	__type(value, struct cell_cpumask_wrapper);
 	__uint(max_entries, MAX_CELLS);
 	__uint(map_flags, 0);
-} cell_cpumasks SEC(".maps");
+} cell_cpumasks			    SEC(".maps");
 
 static inline const struct cpumask *lookup_cell_cpumask(int idx)
 {
@@ -288,12 +380,12 @@ static void cstat_inc(enum cell_stat_idx idx, u32 cell, struct cpu_ctx *cctx)
 }
 
 static inline int update_task_cpumask(struct task_struct *p,
-				      struct task_ctx *tctx)
+				      struct task_ctx	 *tctx)
 {
 	const struct cpumask *cell_cpumask;
-	struct cpu_ctx *cpu_ctx;
-	struct cell *cell;
-	u32 cpu;
+	struct cpu_ctx	     *cpu_ctx;
+	struct cell	     *cell;
+	u32		      cpu;
 
 	if (!(cell_cpumask = lookup_cell_cpumask(tctx->cell)))
 		return -ENOENT;
@@ -326,7 +418,7 @@ static inline int update_task_cpumask(struct task_struct *p,
 		cpu = bpf_cpumask_any_distribute(p->cpus_ptr);
 		if (!(cpu_ctx = lookup_cpu_ctx(cpu)))
 			return -ENOENT;
-		tctx->dsq = cpu_dsq(cpu);
+		tctx->dsq	 = cpu_dsq(cpu);
 		p->scx.dsq_vtime = READ_ONCE(cpu_ctx->vtime_now);
 	}
 
@@ -342,8 +434,35 @@ static inline int update_task_cell(struct task_struct *p, struct task_ctx *tctx,
 {
 	struct cgrp_ctx *cgc;
 
-	if (!(cgc = lookup_cgrp_ctx(cg)))
-		return -ENOENT;
+	cgc = lookup_cgrp_ctx_fallible(cg);
+
+	if (!cgc) {
+		/*
+		 * Cgroup lookup failed - this can happen during scheduler load
+		 * for tasks that were forked before the scheduler was loaded,
+		 * whose cgroups went offline before scx_cgroup_init() ran.
+		 * Only fall back to root cgroup if the workaround is enabled
+		 * and the task is exiting.
+		 */
+		if (exiting_task_workaround_enabled &&
+		    (p->flags & PF_EXITING)) {
+			struct cgroup *rootcg = READ_ONCE(root_cgrp);
+			if (!rootcg) {
+				scx_bpf_error(
+					"Unexpected uninitialized rootcg");
+				return -ENOENT;
+			}
+
+			cgc = lookup_cgrp_ctx(rootcg);
+		}
+
+		if (!cgc) {
+			scx_bpf_error(
+				"cgrp_ctx lookup failed for cgid %llu (task %d, flags 0x%x)",
+				cg->kn->id, p->pid, p->flags);
+			return -ENOENT;
+		}
+	}
 
 	/*
 	 * This ordering is pretty important, we read applied_configuration_seq
@@ -359,12 +478,12 @@ static inline int update_task_cell(struct task_struct *p, struct task_ctx *tctx,
 }
 
 /* Helper function for picking an idle cpu out of a candidate set */
-static s32 pick_idle_cpu_from(struct task_struct *p,
+static s32 pick_idle_cpu_from(struct task_struct   *p,
 			      const struct cpumask *cand_cpumask, s32 prev_cpu,
 			      const struct cpumask *idle_smtmask)
 {
 	bool prev_in_cand = bpf_cpumask_test_cpu(prev_cpu, cand_cpumask);
-	s32 cpu;
+	s32  cpu;
 
 	/*
 	 * If CPU has SMT, any wholly idle CPU is likely a better pick than
@@ -389,10 +508,10 @@ static s32 pick_idle_cpu_from(struct task_struct *p,
 
 /* Check if we need to update the cell/cpumask mapping */
 static __always_inline int maybe_refresh_cell(struct task_struct *p,
-					      struct task_ctx *tctx)
+					      struct task_ctx	 *tctx)
 {
 	struct cgroup *cgrp;
-	int ret = 0;
+	int	       ret = 0;
 	if (tctx->configuration_seq != READ_ONCE(applied_configuration_seq)) {
 		if (!(cgrp = task_cgroup(p)))
 			return -1;
@@ -404,12 +523,12 @@ static __always_inline int maybe_refresh_cell(struct task_struct *p,
 }
 
 static __always_inline s32 pick_idle_cpu(struct task_struct *p, s32 prev_cpu,
-					 struct cpu_ctx *cctx,
+					 struct cpu_ctx	 *cctx,
 					 struct task_ctx *tctx)
 {
-	struct cpumask *task_cpumask;
+	struct cpumask	     *task_cpumask;
 	const struct cpumask *idle_smtmask;
-	s32 cpu;
+	s32		      cpu;
 
 	if (!(task_cpumask = (struct cpumask *)tctx->cpumask) ||
 	    !(idle_smtmask = scx_bpf_get_idle_smtmask())) {
@@ -438,8 +557,8 @@ out:
 s32 BPF_STRUCT_OPS(mitosis_select_cpu, struct task_struct *p, s32 prev_cpu,
 		   u64 wake_flags)
 {
-	s32 cpu;
-	struct cpu_ctx *cctx;
+	s32		 cpu;
+	struct cpu_ctx	*cctx;
 	struct task_ctx *tctx;
 
 	if (!(cctx = lookup_cpu_ctx(-1)) || !(tctx = lookup_task_ctx(p)))
@@ -481,13 +600,13 @@ s32 BPF_STRUCT_OPS(mitosis_select_cpu, struct task_struct *p, s32 prev_cpu,
 
 void BPF_STRUCT_OPS(mitosis_enqueue, struct task_struct *p, u64 enq_flags)
 {
-	struct cpu_ctx *cctx;
+	struct cpu_ctx	*cctx;
 	struct task_ctx *tctx;
-	struct cell *cell;
-	s32 task_cpu = scx_bpf_task_cpu(p);
-	u64 vtime = p->scx.dsq_vtime;
-	s32 cpu = -1;
-	u64 basis_vtime;
+	struct cell	*cell;
+	s32		 task_cpu = scx_bpf_task_cpu(p);
+	u64		 vtime;
+	s32		 cpu = -1;
+	u64		 basis_vtime;
 
 	if (!(tctx = lookup_task_ctx(p)) || !(cctx = lookup_cpu_ctx(-1)))
 		return;
@@ -495,6 +614,9 @@ void BPF_STRUCT_OPS(mitosis_enqueue, struct task_struct *p, u64 enq_flags)
 	if (maybe_refresh_cell(p, tctx) < 0)
 		return;
 
+	/* Ensure this is done *AFTER* refreshing cell which might manipulate vtime */
+	vtime = p->scx.dsq_vtime;
+
 	if (!tctx->all_cell_cpus_allowed) {
 		cpu = dsq_to_cpu(tctx->dsq);
 	} else if (!__COMPAT_is_enq_cpu_selected(enq_flags)) {
@@ -541,7 +663,10 @@ void BPF_STRUCT_OPS(mitosis_enqueue, struct task_struct *p, u64 enq_flags)
 	tctx->basis_vtime = basis_vtime;
 
 	if (time_after(vtime, basis_vtime + 8192 * slice_ns)) {
-		scx_bpf_error("vtime is too far in the future for %d", p->pid);
+		scx_bpf_error(
+			"vtime too far ahead: pid=%d vtime=%llu basis=%llu diff=%llu cell=%u",
+			p->pid, p->scx.dsq_vtime, basis_vtime,
+			p->scx.dsq_vtime - basis_vtime, tctx->cell);
 		return;
 	}
 	/*
@@ -561,31 +686,31 @@ void BPF_STRUCT_OPS(mitosis_enqueue, struct task_struct *p, u64 enq_flags)
 void BPF_STRUCT_OPS(mitosis_dispatch, s32 cpu, struct task_struct *prev)
 {
 	struct cpu_ctx *cctx;
-	u32 cell;
+	u32		cell;
 
 	if (!(cctx = lookup_cpu_ctx(-1)))
 		return;
 
-	cell = READ_ONCE(cctx->cell);
+	cell			  = READ_ONCE(cctx->cell);
 
-	bool found = false;
-	u64 min_vtime_dsq;
-	u64 min_vtime;
+	bool		    found = false;
+	u64		    min_vtime_dsq;
+	u64		    min_vtime;
 
 	struct task_struct *p;
 	bpf_for_each(scx_dsq, p, cell, 0) {
-		min_vtime = p->scx.dsq_vtime;
+		min_vtime     = p->scx.dsq_vtime;
 		min_vtime_dsq = cell;
-		found = true;
+		found	      = true;
 		break;
 	}
 
 	u64 dsq = cpu_dsq(cpu);
 	bpf_for_each(scx_dsq, p, dsq, 0) {
 		if (!found || time_before(p->scx.dsq_vtime, min_vtime)) {
-			min_vtime = p->scx.dsq_vtime;
+			min_vtime     = p->scx.dsq_vtime;
 			min_vtime_dsq = dsq;
-			found = true;
+			found	      = true;
 		}
 		break;
 	}
@@ -630,7 +755,7 @@ void BPF_STRUCT_OPS(mitosis_dispatch, s32 cpu, struct task_struct *prev)
 
 struct cpumask_entry {
 	unsigned long cpumask[CPUMASK_LONG_ENTRIES];
-	u64 used;
+	u64	      used;
 };
 
 struct {
@@ -638,7 +763,7 @@ struct {
 	__type(key, u32);
 	__type(value, struct cpumask_entry);
 	__uint(max_entries, MAX_CPUMASK_ENTRIES);
-} cgrp_init_percpu_cpumask SEC(".maps");
+} cgrp_init_percpu_cpumask	    SEC(".maps");
 
 static inline struct cpumask_entry *allocate_cpumask_entry()
 {
@@ -693,7 +818,7 @@ struct cpuset___cpumask_arr {
  * Given a cgroup, get its cpumask (populated in entry), returns 0 if no
  * cpumask, < 0 on error and > 0 on a populated cpumask.
  */
-static inline int get_cgroup_cpumask(struct cgroup *cgrp,
+static inline int get_cgroup_cpumask(struct cgroup	  *cgrp,
 				     struct cpumask_entry *entry)
 {
 	if (!cgrp->subsys[cpuset_cgrp_id])
@@ -756,39 +881,33 @@ static inline int get_cgroup_cpumask(struct cgroup *cgrp,
  * cgroup hierarchy.
  */
 u32 level_cells[MAX_CG_DEPTH];
-int running;
 
 /*
  * On tick, we identify new cells and apply CPU assignment
  */
-void BPF_STRUCT_OPS(mitosis_tick, struct task_struct *p_run)
+static int update_timer_cb(void *map, int *key, struct bpf_timer *timer)
 {
-	/*
-	 * We serialize tick() on core 0 and ensure only one tick running at a time
-	 * to ensure this can only happen once.
-	 */
-	if (bpf_get_smp_processor_id())
-		return;
+	int ret;
+	if ((ret = bpf_timer_start(timer, TIMER_INTERVAL_NS, 0))) {
+		scx_bpf_error("Failed to arm update timer: %d", ret);
+		return 0;
+	}
 
 	u32 local_configuration_seq = READ_ONCE(configuration_seq);
 	if (local_configuration_seq == READ_ONCE(applied_configuration_seq))
-		return;
-
-	int zero = 0;
-	if (!__atomic_compare_exchange_n(&running, &zero, 1, false,
-					 __ATOMIC_SEQ_CST, __ATOMIC_SEQ_CST))
-		return;
+		return 0;
 
 	DECLARE_CPUMASK_ENTRY(entry) = allocate_cpumask_entry();
 	if (!entry)
-		return;
+		return 0;
 
 	/* Get the root cell (cell 0) and its cpumask */
+	int			     zero = 0;
 	struct cell_cpumask_wrapper *root_cell_cpumaskw;
 	if (!(root_cell_cpumaskw =
 		      bpf_map_lookup_elem(&cell_cpumasks, &zero))) {
 		scx_bpf_error("Failed to find root cell cpumask");
-		return;
+		return 0;
 	}
 
 	struct bpf_cpumask *root_bpf_cpumask;
@@ -796,16 +915,17 @@ void BPF_STRUCT_OPS(mitosis_tick, struct task_struct *p_run)
 		bpf_kptr_xchg(&root_cell_cpumaskw->tmp_cpumask, NULL);
 	if (!root_bpf_cpumask) {
 		scx_bpf_error("tmp_cpumask should never be null");
-		return;
+		return 0;
 	}
 	if (!root_cell_cpumaskw->cpumask) {
 		scx_bpf_error("root cpumasks should never be null");
 		goto out;
 	}
 
+	bpf_rcu_read_lock();
 	if (!all_cpumask) {
 		scx_bpf_error("NULL all_cpumask");
-		goto out;
+		goto out_rcu_unlock;
 	}
 
 	/*
@@ -815,29 +935,28 @@ void BPF_STRUCT_OPS(mitosis_tick, struct task_struct *p_run)
 	bpf_cpumask_copy(root_bpf_cpumask, (const struct cpumask *)all_cpumask);
 
 	struct cgroup_subsys_state *root_css, *pos;
-	struct cgroup *cur_cgrp, *root_cgrp_ref;
+	struct cgroup		   *cur_cgrp, *root_cgrp_ref;
 
 	if (!root_cgrp) {
 		scx_bpf_error("root_cgrp should not be null");
-		goto out;
+		goto out_rcu_unlock;
 	}
 
 	struct cgrp_ctx *root_cgrp_ctx;
 	if (!(root_cgrp_ctx = lookup_cgrp_ctx(root_cgrp)))
-		goto out;
+		goto out_rcu_unlock;
 
 	if (!root_cgrp) {
 		scx_bpf_error("root_cgrp should not be null");
-		goto out;
+		goto out_rcu_unlock;
 	}
 
 	if (!(root_cgrp_ref = bpf_cgroup_acquire(root_cgrp))) {
 		scx_bpf_error("Failed to acquire reference to root_cgrp");
-		goto out;
+		goto out_rcu_unlock;
 	}
 	root_css = &root_cgrp_ref->self;
 
-	bpf_rcu_read_lock();
 	/*
 	 * Iterate over all cgroups, check if any have a cpumask and populate them
 	 * as a separate cell.
@@ -867,7 +986,7 @@ void BPF_STRUCT_OPS(mitosis_tick, struct task_struct *p_run)
 					scx_bpf_error(
 						"Cgroup hierarchy is too deep: %d",
 						level);
-					goto out_rcu_unlock;
+					goto out_root_cgrp;
 				}
 				/*
 				 * This is a janky way of getting the parent cell, ideally we'd
@@ -889,7 +1008,7 @@ void BPF_STRUCT_OPS(mitosis_tick, struct task_struct *p_run)
 			}
 			continue;
 		} else if (rc < 0)
-			goto out_rcu_unlock;
+			goto out_root_cgrp;
 
 		/*
 		 * cgroup has a cpumask, allocate a new cell if needed, and assign cpus
@@ -898,7 +1017,7 @@ void BPF_STRUCT_OPS(mitosis_tick, struct task_struct *p_run)
 		if (!cgrp_ctx->cell_owner) {
 			cell_idx = allocate_cell();
 			if (cell_idx < 0)
-				goto out_rcu_unlock;
+				goto out_root_cgrp;
 			cgrp_ctx->cell_owner = true;
 		}
 
@@ -907,14 +1026,14 @@ void BPF_STRUCT_OPS(mitosis_tick, struct task_struct *p_run)
 			      bpf_map_lookup_elem(&cell_cpumasks, &cell_idx))) {
 			scx_bpf_error("Failed to find cell cpumask: %d",
 				      cell_idx);
-			goto out_rcu_unlock;
+			goto out_root_cgrp;
 		}
 
 		struct bpf_cpumask *bpf_cpumask;
 		bpf_cpumask = bpf_kptr_xchg(&cell_cpumaskw->tmp_cpumask, NULL);
 		if (!bpf_cpumask) {
 			scx_bpf_error("tmp_cpumask should never be null");
-			goto out_rcu_unlock;
+			goto out_root_cgrp;
 		}
 		bpf_cpumask_copy(bpf_cpumask,
 				 (const struct cpumask *)&entry->cpumask);
@@ -927,7 +1046,7 @@ void BPF_STRUCT_OPS(mitosis_tick, struct task_struct *p_run)
 				struct cpu_ctx *cpu_ctx;
 				if (!(cpu_ctx = lookup_cpu_ctx(cpu_idx))) {
 					bpf_cpumask_release(bpf_cpumask);
-					goto out_rcu_unlock;
+					goto out_root_cgrp;
 				}
 				cpu_ctx->cell = cell_idx;
 				bpf_cpumask_clear_cpu(cpu_idx,
@@ -938,7 +1057,7 @@ void BPF_STRUCT_OPS(mitosis_tick, struct task_struct *p_run)
 			bpf_kptr_xchg(&cell_cpumaskw->cpumask, bpf_cpumask);
 		if (!bpf_cpumask) {
 			scx_bpf_error("cpumask should never be null");
-			goto out_rcu_unlock;
+			goto out_root_cgrp;
 		}
 
 		bpf_cpumask =
@@ -946,7 +1065,7 @@ void BPF_STRUCT_OPS(mitosis_tick, struct task_struct *p_run)
 		if (bpf_cpumask) {
 			scx_bpf_error("tmp_cpumask should be null");
 			bpf_cpumask_release(bpf_cpumask);
-			goto out_rcu_unlock;
+			goto out_root_cgrp;
 		}
 
 		barrier();
@@ -955,11 +1074,10 @@ void BPF_STRUCT_OPS(mitosis_tick, struct task_struct *p_run)
 		if (level <= 0 || level >= MAX_CG_DEPTH) {
 			scx_bpf_error("Cgroup hierarchy is too deep: %d",
 				      level);
-			goto out_rcu_unlock;
+			goto out_root_cgrp;
 		}
 		level_cells[level] = cell_idx;
 	}
-	bpf_rcu_read_unlock();
 
 	/*
 	 * assign root cell cpus that are left over
@@ -967,13 +1085,12 @@ void BPF_STRUCT_OPS(mitosis_tick, struct task_struct *p_run)
 	int cpu_idx;
 	bpf_for(cpu_idx, 0, nr_possible_cpus)
 	{
-		if (bpf_cpumask_test_cpu(
-			    cpu_idx, (const struct cpumask *)&entry->cpumask)) {
+		if (bpf_cpumask_test_cpu(cpu_idx, (const struct cpumask *)
+							  root_bpf_cpumask)) {
 			struct cpu_ctx *cpu_ctx;
 			if (!(cpu_ctx = lookup_cpu_ctx(cpu_idx)))
 				goto out_root_cgrp;
 			cpu_ctx->cell = 0;
-			bpf_cpumask_clear_cpu(cpu_idx, root_bpf_cpumask);
 		}
 	}
 
@@ -981,8 +1098,9 @@ void BPF_STRUCT_OPS(mitosis_tick, struct task_struct *p_run)
 		bpf_kptr_xchg(&root_cell_cpumaskw->cpumask, root_bpf_cpumask);
 	if (!root_bpf_cpumask) {
 		scx_bpf_error("root cpumask should never be null");
+		bpf_rcu_read_unlock();
 		bpf_cgroup_release(root_cgrp_ref);
-		return;
+		return 0;
 	}
 
 	root_bpf_cpumask = bpf_kptr_xchg(&root_cell_cpumaskw->tmp_cpumask,
@@ -994,62 +1112,89 @@ void BPF_STRUCT_OPS(mitosis_tick, struct task_struct *p_run)
 
 	barrier();
 	WRITE_ONCE(applied_configuration_seq, local_configuration_seq);
-	WRITE_ONCE(running, 0);
 
-	bpf_cgroup_release(root_cgrp_ref);
-	return;
-out_rcu_unlock:
 	bpf_rcu_read_unlock();
+	bpf_cgroup_release(root_cgrp_ref);
+	return 0;
 out_root_cgrp:
 	bpf_cgroup_release(root_cgrp_ref);
+out_rcu_unlock:
+	bpf_rcu_read_unlock();
 out:
 	bpf_cpumask_release(root_bpf_cpumask);
+	return 0;
 }
 
 void BPF_STRUCT_OPS(mitosis_running, struct task_struct *p)
 {
-	struct cpu_ctx *cctx;
+	struct cpu_ctx	*cctx;
 	struct task_ctx *tctx;
-	struct cell *cell;
+	struct cell	*cell;
 
-	if (!(tctx = lookup_task_ctx(p)) || !(cctx = lookup_cpu_ctx(-1)) ||
-	    !(cell = lookup_cell(cctx->cell)))
+	if (!(tctx = lookup_task_ctx(p)))
 		return;
 
 	/*
-	 * Update both the CPU's cell and the cpu's vtime so the vtime's are
-	 * comparable at dispatch time.
+	 * Legacy approach: Update vtime_now before task runs.
+	 * Only used when split vtime updates is enabled.
 	 */
-	if (time_before(READ_ONCE(cell->vtime_now), p->scx.dsq_vtime))
-		WRITE_ONCE(cell->vtime_now, p->scx.dsq_vtime);
+	if (split_vtime_updates) {
+		if (!(cctx = lookup_cpu_ctx(-1)) ||
+		    !(cell = lookup_cell(cctx->cell)))
+			return;
 
-	if (time_before(READ_ONCE(cctx->vtime_now), p->scx.dsq_vtime))
-		WRITE_ONCE(cctx->vtime_now, p->scx.dsq_vtime);
+		if (time_before(READ_ONCE(cell->vtime_now), p->scx.dsq_vtime))
+			WRITE_ONCE(cell->vtime_now, p->scx.dsq_vtime);
+
+		if (time_before(READ_ONCE(cctx->vtime_now), p->scx.dsq_vtime))
+			WRITE_ONCE(cctx->vtime_now, p->scx.dsq_vtime);
+	}
 
 	tctx->started_running_at = scx_bpf_now();
 }
 
 void BPF_STRUCT_OPS(mitosis_stopping, struct task_struct *p, bool runnable)
 {
-	struct cpu_ctx *cctx;
+	struct cpu_ctx	*cctx;
 	struct task_ctx *tctx;
-	struct cell *cell;
-	u64 now, used;
-	u32 cidx;
+	struct cell	*cell;
+	u64		 now, used;
+	u32		 cidx;
 
 	if (!(cctx = lookup_cpu_ctx(-1)) || !(tctx = lookup_task_ctx(p)))
 		return;
 
-	cidx = tctx->cell;
+	/*
+	 * Use CPU's cell (not task's cell) to match dispatch() logic.
+	 * Prevents starvation when a task is pinned outside its cell.
+	 * E.g. a cell 0 kworker pinned to a cell 1 CPU.
+	 */
+	cidx = cctx->cell;
 	if (!(cell = lookup_cell(cidx)))
 		return;
 
-	now = scx_bpf_now();
-	used = now - tctx->started_running_at;
+	now			 = scx_bpf_now();
+	used			 = now - tctx->started_running_at;
 	tctx->started_running_at = now;
 	/* scale the execution time by the inverse of the weight and charge */
+	if (p->scx.weight == 0) {
+		scx_bpf_error("Task %d has zero weight", p->pid);
+		return;
+	}
 	p->scx.dsq_vtime += used * 100 / p->scx.weight;
 
+	/*
+	 * Default approach: Update cell and cpu dsq vtime after updating task's vtime
+	 * to keep them in sync and prevent "vtime too far ahead" errors.
+	 */
+	if (!split_vtime_updates) {
+		if (time_before(READ_ONCE(cell->vtime_now), p->scx.dsq_vtime))
+			WRITE_ONCE(cell->vtime_now, p->scx.dsq_vtime);
+
+		if (time_before(READ_ONCE(cctx->vtime_now), p->scx.dsq_vtime))
+			WRITE_ONCE(cctx->vtime_now, p->scx.dsq_vtime);
+	}
+
 	if (cidx != 0 || tctx->all_cell_cpus_allowed) {
 		u64 *cell_cycles = MEMBER_VPTR(cctx->cell_cycles, [cidx]);
 		if (!cell_cycles) {
@@ -1076,6 +1221,9 @@ s32 BPF_STRUCT_OPS(mitosis_cgroup_init, struct cgroup *cgrp,
 		   struct scx_cgroup_init_args *args)
 {
 	struct cgrp_ctx *cgc;
+
+	record_cgroup_init(cgrp->kn->id);
+
 	if (!(cgc = bpf_cgrp_storage_get(&cgrp_ctxs, cgrp, 0,
 					 BPF_LOCAL_STORAGE_GET_F_CREATE))) {
 		scx_bpf_error("cgrp_ctx creation failed for cgid %llu",
@@ -1121,15 +1269,20 @@ s32 BPF_STRUCT_OPS(mitosis_cgroup_init, struct cgroup *cgrp,
 s32 BPF_STRUCT_OPS(mitosis_cgroup_exit, struct cgroup *cgrp)
 {
 	struct cgrp_ctx *cgc;
-	if (!(cgc = bpf_cgrp_storage_get(&cgrp_ctxs, cgrp, 0,
-					 BPF_LOCAL_STORAGE_GET_F_CREATE))) {
-		scx_bpf_error("cgrp_ctx creation failed for cgid %llu",
-			      cgrp->kn->id);
-		return -ENOENT;
+	int		 ret;
+
+	record_cgroup_exit(cgrp->kn->id);
+
+	/*
+	 * Use lookup without CREATE since this is the exit path. If the cgroup
+	 * doesn't have storage, it's not a cell owner anyway.
+	 */
+	if (!(cgc = lookup_cgrp_ctx(cgrp))) {
+		/* Errors above on failure, verifier. */
+		return 0;
 	}
 
 	if (cgc->cell_owner) {
-		int ret;
 		if ((ret = free_cell(cgc->cell)))
 			return ret;
 		/*
@@ -1173,9 +1326,11 @@ void BPF_STRUCT_OPS(mitosis_set_cpumask, struct task_struct *p,
 s32 BPF_STRUCT_OPS(mitosis_init_task, struct task_struct *p,
 		   struct scx_init_task_args *args)
 {
-	struct task_ctx *tctx;
+	struct task_ctx	   *tctx;
 	struct bpf_cpumask *cpumask;
 
+	record_init_task(args->cgroup->kn->id, p->pid);
+
 	tctx = bpf_task_storage_get(&task_ctxs, p, 0,
 				    BPF_LOCAL_STORAGE_GET_F_CREATE);
 	if (!tctx) {
@@ -1241,9 +1396,9 @@ static void dump_cell_cpumask(int id)
 
 void BPF_STRUCT_OPS(mitosis_dump, struct scx_dump_ctx *dctx)
 {
-	u64 dsq_id;
-	int i;
-	struct cell *cell;
+	u64		dsq_id;
+	int		i;
+	struct cell    *cell;
 	struct cpu_ctx *cpu_ctx;
 
 	scx_bpf_dump_header();
@@ -1274,6 +1429,55 @@ void BPF_STRUCT_OPS(mitosis_dump, struct scx_dump_ctx *dctx)
 			     cpu_ctx->cell, READ_ONCE(cpu_ctx->vtime_now),
 			     scx_bpf_dsq_nr_queued(dsq_id));
 	}
+
+	if (!debug_events_enabled)
+		return;
+
+	/* Dump debug events */
+	scx_bpf_dump("\n");
+	scx_bpf_dump("DEBUG EVENTS (last %d):\n", DEBUG_EVENTS_BUF_SIZE);
+
+	u32 total_events = READ_ONCE(debug_event_pos);
+	u32 start_idx	 = total_events > DEBUG_EVENTS_BUF_SIZE ?
+				   total_events - DEBUG_EVENTS_BUF_SIZE :
+				   0;
+
+	bpf_for(i, 0, DEBUG_EVENTS_BUF_SIZE)
+	{
+		u32 event_num = start_idx + i;
+		if (event_num >= total_events)
+			break;
+
+		u32		    idx = event_num % DEBUG_EVENTS_BUF_SIZE;
+		struct debug_event *event =
+			bpf_map_lookup_elem(&debug_events, &idx);
+		if (!event)
+			continue;
+
+		switch (event->event_type) {
+		case DEBUG_EVENT_CGROUP_INIT:
+			scx_bpf_dump("[%3d] CGROUP_INIT cgid=%llu ts=%llu\n",
+				     event_num, event->cgroup_init.cgid,
+				     event->timestamp);
+			break;
+		case DEBUG_EVENT_INIT_TASK:
+			scx_bpf_dump(
+				"[%3d] INIT_TASK   cgid=%llu pid=%u ts=%llu\n",
+				event_num, event->init_task.cgid,
+				event->init_task.pid, event->timestamp);
+			break;
+		case DEBUG_EVENT_CGROUP_EXIT:
+			scx_bpf_dump("[%3d] CGROUP_EXIT cgid=%llu ts=%llu\n",
+				     event_num, event->cgroup_exit.cgid,
+				     event->timestamp);
+			break;
+		default:
+			scx_bpf_dump("[%3d] UNKNOWN     type=%u ts=%llu\n",
+				     event_num, event->event_type,
+				     event->timestamp);
+			break;
+		}
+	}
 }
 
 void BPF_STRUCT_OPS(mitosis_dump_task, struct scx_dump_ctx *dctx,
@@ -1296,13 +1500,23 @@ void BPF_STRUCT_OPS(mitosis_dump_task, struct scx_dump_ctx *dctx,
 s32 BPF_STRUCT_OPS_SLEEPABLE(mitosis_init)
 {
 	struct bpf_cpumask *cpumask;
-	u32 i;
-	s32 ret;
+	u32		    i;
+	s32		    ret;
 
-	struct cgroup *rootcg;
+	u32		    key = 0;
+
+	struct cgroup	   *rootcg;
 	if (!(rootcg = bpf_cgroup_from_id(root_cgid)))
 		return -ENOENT;
 
+	/* initialize cgrp storage for rootcg so that it is always available in the timer */
+	if (!bpf_cgrp_storage_get(&cgrp_ctxs, rootcg, 0,
+				  BPF_LOCAL_STORAGE_GET_F_CREATE)) {
+		scx_bpf_error("cgrp_ctx creation failed for rootcg");
+		bpf_cgroup_release(rootcg);
+		return -ENOENT;
+	}
+
 	rootcg = bpf_kptr_xchg(&root_cgrp, rootcg);
 	if (rootcg)
 		bpf_cgroup_release(rootcg);
@@ -1373,7 +1587,18 @@ s32 BPF_STRUCT_OPS_SLEEPABLE(mitosis_init)
 		}
 	}
 
-	cells[0].in_use = true;
+	cells[0].in_use		= true;
+	struct bpf_timer *timer = bpf_map_lookup_elem(&update_timer, &key);
+	if (!timer) {
+		scx_bpf_error("Failed to lookup update timer");
+		return -ESRCH;
+	}
+	bpf_timer_init(timer, &update_timer, CLOCK_BOOTTIME);
+	bpf_timer_set_callback(timer, update_timer_cb);
+	if ((ret = bpf_timer_start(timer, TIMER_INTERVAL_NS, 0))) {
+		scx_bpf_error("Failed to arm update timer");
+		return ret;
+	}
 	return 0;
 }
 
@@ -1382,22 +1607,21 @@ void BPF_STRUCT_OPS(mitosis_exit, struct scx_exit_info *ei)
 	UEI_RECORD(uei, ei);
 }
 
-SEC(".struct_ops.link")
-struct sched_ext_ops mitosis = {
-	.select_cpu = (void *)mitosis_select_cpu,
-	.enqueue = (void *)mitosis_enqueue,
-	.dispatch = (void *)mitosis_dispatch,
-	.tick = (void *)mitosis_tick,
-	.running = (void *)mitosis_running,
-	.stopping = (void *)mitosis_stopping,
-	.set_cpumask = (void *)mitosis_set_cpumask,
-	.init_task = (void *)mitosis_init_task,
-	.cgroup_init = (void *)mitosis_cgroup_init,
-	.cgroup_exit = (void *)mitosis_cgroup_exit,
-	.cgroup_move = (void *)mitosis_cgroup_move,
-	.dump = (void *)mitosis_dump,
-	.dump_task = (void *)mitosis_dump_task,
-	.init = (void *)mitosis_init,
-	.exit = (void *)mitosis_exit,
-	.name = "mitosis",
-};
+// clang-format off
+SCX_OPS_DEFINE(mitosis,
+	       .select_cpu		= (void *)mitosis_select_cpu,
+	       .enqueue			= (void *)mitosis_enqueue,
+	       .dispatch		= (void *)mitosis_dispatch,
+	       .running			= (void *)mitosis_running,
+	       .stopping		= (void *)mitosis_stopping,
+	       .set_cpumask		= (void *)mitosis_set_cpumask,
+	       .init_task		= (void *)mitosis_init_task,
+	       .cgroup_init		= (void *)mitosis_cgroup_init,
+	       .cgroup_exit		= (void *)mitosis_cgroup_exit,
+	       .cgroup_move		= (void *)mitosis_cgroup_move,
+	       .dump 			= (void *)mitosis_dump,
+	       .dump_task		= (void *)mitosis_dump_task,
+	       .init			= (void *)mitosis_init,
+	       .exit			= (void *)mitosis_exit,
+	       .name			= "mitosis");
+// clang-format on
