@@ -6,6 +6,7 @@ pub mod bpf_intf;
 pub mod bpf_skel;
 pub use bpf_skel::types;
 
+use scx_utils::cli::TopologyArgs;
 pub use scx_utils::CoreType;
 use scx_utils::Topology;
 pub use scx_utils::NR_CPU_IDS;
@@ -27,6 +28,19 @@ fn get_default_llc_runs() -> u64 {
     llc_runs as u64
 }
 
+fn get_default_llc_shards() -> u32 {
+    let max_cpus_per_llc = TOPO
+        .all_llcs
+        .values()
+        .map(|l| l.all_cpus.len() as u32)
+        .max()
+        .unwrap_or(1);
+    if max_cpus_per_llc <= 4 {
+        return 0;
+    }
+    max_cpus_per_llc / 4
+}
+
 #[derive(Debug, Clone, PartialEq, Eq, PartialOrd, Ord, ValueEnum)]
 pub enum LbMode {
     /// load of the LLC
@@ -64,7 +78,70 @@ impl SchedMode {
     }
 }
 
-#[derive(Debug, Parser)]
+#[derive(Debug, Clone)]
+pub struct HardwareProfile {
+    pub single_llc: bool,
+    pub core_count: usize,
+    pub is_arm64: bool,
+    pub is_neoverse_v2: bool,
+}
+
+impl HardwareProfile {
+    pub fn detect() -> Self {
+        let single_llc = TOPO.all_llcs.len() == 1;
+        let core_count = TOPO.all_cpus.len();
+
+        let is_arm64 = cfg!(target_arch = "aarch64");
+        let is_neoverse_v2 = is_arm64 && Self::detect_neoverse_v2();
+
+        Self {
+            single_llc,
+            core_count,
+            is_arm64,
+            is_neoverse_v2,
+        }
+    }
+
+    fn detect_neoverse_v2() -> bool {
+        // Read /proc/cpuinfo to detect Neoverse-V2
+        if let Ok(cpuinfo) = std::fs::read_to_string("/proc/cpuinfo") {
+            cpuinfo.contains("Neoverse-V2")
+        } else {
+            false
+        }
+    }
+
+    pub fn optimize_scheduler_opts(&self, opts: &mut SchedulerOpts) {
+        if self.single_llc {
+            // Disable pick2 dispatch since there's only one LLC
+            opts.dispatch_pick2_disable = true;
+            // Disable wakeup LLC migrations
+            opts.wakeup_llc_migrations = false;
+            // For single LLC, set min_llc_runs_pick2 to 0 (effectively disable)
+            opts.min_llc_runs_pick2 = 0;
+            log::info!("Single LLC detected - disabling multi-LLC optimizations");
+        }
+
+        if self.is_neoverse_v2 && self.single_llc && self.core_count >= 64 {
+            // Optimize shard count for large single-LLC systems
+            let optimal_shards = (self.core_count / 8).min(16) as u32;
+            if opts.llc_shards == get_default_llc_shards() {
+                opts.llc_shards = optimal_shards;
+                log::info!(
+                    "Large single-LLC system ({} cores) - using {} shards",
+                    self.core_count,
+                    opts.llc_shards
+                );
+            }
+        }
+
+        if self.is_neoverse_v2 {
+            log::info!("ARM64 Neoverse-V2 detected - ARM-specific optimizations available");
+        }
+    }
+}
+
+#[derive(Debug, Clone, Parser)]
 pub struct SchedulerOpts {
     /// Disables per-cpu kthreads directly dispatched into local dsqs.
     #[clap(short = 'k', long, action = clap::ArgAction::SetTrue)]
@@ -198,6 +275,10 @@ pub struct SchedulerOpts {
     #[clap(short = 'x', long, default_value = "4")]
     pub dsq_shift: u64,
 
+    /// Number of shards for LLC DSQ sharding. Default 0 means use single DSQ per LLC.
+    #[clap(long, default_value_t = get_default_llc_shards())]
+    pub llc_shards: u32,
+
     /// Minimum number of queued tasks to use pick2 balancing, 0 to always enabled.
     #[clap(short = 'm', long, default_value = "0")]
     pub min_nr_queued_pick2: u64,
@@ -209,6 +290,21 @@ pub struct SchedulerOpts {
     /// Initial DSQ for tasks.
     #[clap(short = 'i', long, default_value = "0")]
     pub init_dsq_index: usize,
+
+    /// Use a arena based queues (ATQ) for task queueing.
+    #[clap(long, default_value_t = false, action = clap::ArgAction::Set)]
+    pub virt_llc_enabled: bool,
+
+    /// Enable hardware-specific optimizations automatically
+    #[clap(long, default_value_t = true, action = clap::ArgAction::Set)]
+    pub hw_auto_optimize: bool,
+
+    /// Force single-LLC fast path optimizations
+    #[clap(long, action = clap::ArgAction::SetTrue)]
+    pub single_llc_fast_path: bool,
+
+    #[clap(flatten, next_help_heading = "Topology Options")]
+    pub topo: TopologyArgs,
 }
 
 pub fn dsq_slice_ns(dsq_index: u64, min_slice_us: u64, dsq_shift: u64) -> u64 {
@@ -221,11 +317,12 @@ pub fn dsq_slice_ns(dsq_index: u64, min_slice_us: u64, dsq_shift: u64) -> u64 {
 
 #[macro_export]
 macro_rules! init_open_skel {
-    ($skel: expr, $opts: expr, $verbose: expr) => {
+    ($skel: expr, $topo: expr, $opts: expr, $verbose: expr, $hw_profile: expr) => {
         'block: {
-            let skel = $skel;
+            let skel = &mut *$skel;
             let opts: &$crate::SchedulerOpts = $opts;
             let verbose: u8 = $verbose;
+            let hw_profile: &$crate::HardwareProfile = $hw_profile;
 
             if opts.init_dsq_index > opts.dumb_queues - 1 {
                 break 'block ::anyhow::Result::Err(::anyhow::anyhow!(
@@ -272,10 +369,10 @@ macro_rules! init_open_skel {
             // topo config
             let rodata = skel.maps.rodata_data.as_mut().unwrap();
             rodata.topo_config.nr_cpus = *$crate::NR_CPU_IDS as u32;
-            rodata.topo_config.nr_llcs = $crate::TOPO.all_llcs.clone().keys().len() as u32;
-            rodata.topo_config.nr_nodes = $crate::TOPO.nodes.clone().keys().len() as u32;
-            rodata.topo_config.smt_enabled = MaybeUninit::new($crate::TOPO.smt_enabled);
-            rodata.topo_config.has_little_cores = MaybeUninit::new($crate::TOPO.has_little_cores());
+            rodata.topo_config.nr_llcs = $topo.all_llcs.clone().keys().len() as u32;
+            rodata.topo_config.nr_nodes = $topo.nodes.clone().keys().len() as u32;
+            rodata.topo_config.smt_enabled = MaybeUninit::new($topo.smt_enabled);
+            rodata.topo_config.has_little_cores = MaybeUninit::new($topo.has_little_cores());
 
             // timeline config
             rodata.timeline_config.min_slice_us = opts.min_slice_us;
@@ -296,6 +393,9 @@ macro_rules! init_open_skel {
                 MaybeUninit::new(opts.dispatch_lb_interactive);
             rodata.lb_config.wakeup_lb_busy = opts.wakeup_lb_busy;
             rodata.lb_config.wakeup_llc_migrations = MaybeUninit::new(opts.wakeup_llc_migrations);
+            rodata.lb_config.single_llc_mode = MaybeUninit::new(
+                opts.single_llc_fast_path || (opts.hw_auto_optimize && hw_profile.single_llc),
+            );
 
             // p2dq config
             rodata.p2dq_config.interactive_ratio = opts.interactive_ratio as u32;
@@ -306,11 +406,20 @@ macro_rules! init_open_skel {
             rodata.p2dq_config.init_dsq_index = opts.init_dsq_index as i32;
             rodata.p2dq_config.saturated_percent = opts.saturated_percent;
             rodata.p2dq_config.sched_mode = opts.sched_mode.clone() as u32;
+            rodata.p2dq_config.llc_shards = opts.llc_shards.max(1);
 
             rodata.p2dq_config.atq_enabled = MaybeUninit::new(
                 opts.atq_enabled && compat::ksym_exists("bpf_spin_unlock").unwrap_or(false),
             );
-            rodata.p2dq_config.cpu_priority = MaybeUninit::new(opts.cpu_priority);
+
+            // Check if cpu_priority is supported by the kernel
+            let cpu_priority_supported = compat::ksym_exists("sched_core_priority").unwrap_or(false);
+            if opts.cpu_priority && !cpu_priority_supported {
+                ::log::warn!("CPU priority scheduling requested but kernel doesn't support sched_core_priority");
+                ::log::warn!("Feature disabled - requires kernel with hybrid CPU support and appropriate CPU governor (e.g., amd-pstate)");
+                ::log::warn!("See README for kernel requirements");
+            }
+            rodata.p2dq_config.cpu_priority = MaybeUninit::new(opts.cpu_priority && cpu_priority_supported);
             rodata.p2dq_config.freq_control = MaybeUninit::new(opts.freq_control);
             rodata.p2dq_config.interactive_sticky = MaybeUninit::new(opts.interactive_sticky);
             rodata.p2dq_config.keep_running_enabled = MaybeUninit::new(opts.keep_running);
@@ -325,18 +434,19 @@ macro_rules! init_open_skel {
 
 #[macro_export]
 macro_rules! init_skel {
-    ($skel: expr) => {
-        for cpu in $crate::TOPO.all_cpus.values() {
+    ($skel: expr, $topo: expr) => {
+        for cpu in $topo.all_cpus.values() {
             $skel.maps.bss_data.as_mut().unwrap().big_core_ids[cpu.id] =
                 if cpu.core_type == ($crate::CoreType::Big { turbo: true }) {
                     1
                 } else {
                     0
                 };
+            $skel.maps.bss_data.as_mut().unwrap().cpu_core_ids[cpu.id] = cpu.core_id as u32;
             $skel.maps.bss_data.as_mut().unwrap().cpu_llc_ids[cpu.id] = cpu.llc_id as u64;
             $skel.maps.bss_data.as_mut().unwrap().cpu_node_ids[cpu.id] = cpu.node_id as u64;
         }
-        for llc in $crate::TOPO.all_llcs.values() {
+        for llc in $topo.all_llcs.values() {
             $skel.maps.bss_data.as_mut().unwrap().llc_ids[llc.id] = llc.id as u64;
         }
     };
