@@ -16,10 +16,12 @@ use anyhow::Context;
 use anyhow::Result;
 use clap::Parser;
 use crossbeam::channel::RecvTimeoutError;
+use libbpf_rs::skel::Skel;
 use libbpf_rs::MapCore as _;
 use libbpf_rs::OpenObject;
 use libbpf_rs::ProgramInput;
 use log::{debug, info, warn};
+use scx_arena::ArenaLib;
 use scx_stats::prelude::*;
 use scx_utils::build_id;
 use scx_utils::compat;
@@ -34,9 +36,6 @@ use scx_utils::uei_report;
 use scx_utils::Topology;
 use scx_utils::UserExitInfo;
 use scx_utils::NR_CPU_IDS;
-use scx_utils::{Core, Llc};
-
-use std::ffi::c_ulong;
 
 use bpf_intf::stat_idx_P2DQ_NR_STATS;
 use bpf_intf::stat_idx_P2DQ_STAT_ATQ_ENQ;
@@ -118,6 +117,11 @@ impl<'a> Scheduler<'a> {
             "Running scx_p2dq (build ID: {})",
             build_id::full_version(env!("CARGO_PKG_VERSION"))
         );
+        let topo = if opts.virt_llc_enabled {
+            Topology::with_args(&opts.topo)?
+        } else {
+            Topology::new()?
+        };
         let open_opts = libbpf_ops.clone().into_bpf_open_opts();
         let mut open_skel = scx_ops_open!(skel_builder, open_object, p2dq, open_opts).context(
             "Failed to open BPF object. This can be caused by a mismatch between the kernel \
@@ -127,7 +131,15 @@ impl<'a> Scheduler<'a> {
             to this issue or file an issue on the scx repo if the problem persists. \
             https://github.com/sched-ext/scx/issues/new?labels=scx_p2dq&title=scx_p2dq:%20New%20Issue&assignees=hodgesds&body=Kernel%20version:%20(fill%20me%20out)%0ADistribution:%20(fill%20me%20out)%0AHardware:%20(fill%20me%20out)%0A%0AIssue:%20(fill%20me%20out)"
         )?;
-        scx_p2dq::init_open_skel!(&mut open_skel, opts, verbose)?;
+
+        // Apply hardware-specific optimizations before macro
+        let hw_profile = scx_p2dq::HardwareProfile::detect();
+        let mut opts_optimized = opts.clone();
+        if opts.hw_auto_optimize {
+            hw_profile.optimize_scheduler_opts(&mut opts_optimized);
+        }
+
+        scx_p2dq::init_open_skel!(&mut open_skel, topo, &opts_optimized, verbose, &hw_profile)?;
 
         if opts.queued_wakeup {
             open_skel.struct_ops.p2dq_mut().flags |= *compat::SCX_OPS_ALLOW_QUEUED_WAKEUP;
@@ -135,7 +147,7 @@ impl<'a> Scheduler<'a> {
         open_skel.struct_ops.p2dq_mut().flags |= *compat::SCX_OPS_KEEP_BUILTIN_IDLE;
 
         let mut skel = scx_ops_load!(open_skel, p2dq, uei)?;
-        scx_p2dq::init_skel!(&mut skel);
+        scx_p2dq::init_skel!(&mut skel, topo);
 
         let stats_server = StatsServer::new(stats::server_data()).launch()?;
 
@@ -183,126 +195,6 @@ impl<'a> Scheduler<'a> {
         }
     }
 
-    fn setup_arenas(&mut self) -> Result<()> {
-        // Allocate the arena memory from the BPF side so userspace initializes it before starting
-        // the scheduler. Despite the function call's name this is neither a test nor a test run,
-        // it's the recommended way of executing SEC("syscall") probes.
-        let mut args = types::arena_init_args {
-            static_pages: bpf_intf::consts_STATIC_ALLOC_PAGES_GRANULARITY as c_ulong,
-            task_ctx_size: std::mem::size_of::<types::task_p2dq>() as c_ulong,
-        };
-
-        let input = ProgramInput {
-            context_in: Some(unsafe {
-                std::slice::from_raw_parts_mut(
-                    &mut args as *mut _ as *mut u8,
-                    std::mem::size_of_val(&args),
-                )
-            }),
-            ..Default::default()
-        };
-
-        let output = self.skel.progs.arena_init.test_run(input)?;
-        if output.return_value != 0 {
-            bail!(
-                "Could not initialize arenas, p2dq_setup returned {}",
-                output.return_value as i32
-            );
-        }
-
-        Ok(())
-    }
-
-    fn setup_topology_node(&mut self, mask: &[u64]) -> Result<()> {
-        let mut args = types::arena_alloc_mask_args {
-            bitmap: 0 as c_ulong,
-        };
-
-        let input = ProgramInput {
-            context_in: Some(unsafe {
-                std::slice::from_raw_parts_mut(
-                    &mut args as *mut _ as *mut u8,
-                    std::mem::size_of_val(&args),
-                )
-            }),
-            ..Default::default()
-        };
-
-        let output = self.skel.progs.arena_alloc_mask.test_run(input)?;
-        if output.return_value != 0 {
-            bail!(
-                "Could not initialize arenas, setup_topology_node returned {}",
-                output.return_value as i32
-            );
-        }
-
-        let ptr = unsafe { std::mem::transmute::<u64, &mut [u64; 10]>(args.bitmap) };
-
-        let (valid_mask, _) = ptr.split_at_mut(mask.len());
-        valid_mask.clone_from_slice(mask);
-
-        let mut args = types::arena_topology_node_init_args {
-            bitmap: args.bitmap as c_ulong,
-            data_size: 0 as c_ulong,
-            id: 0 as c_ulong,
-        };
-
-        let input = ProgramInput {
-            context_in: Some(unsafe {
-                std::slice::from_raw_parts_mut(
-                    &mut args as *mut _ as *mut u8,
-                    std::mem::size_of_val(&args),
-                )
-            }),
-            ..Default::default()
-        };
-
-        let output = self.skel.progs.arena_topology_node_init.test_run(input)?;
-        if output.return_value != 0 {
-            bail!(
-                "p2dq_topology_node_init returned {}",
-                output.return_value as i32
-            );
-        }
-
-        Ok(())
-    }
-
-    fn setup_topology(&mut self) -> Result<()> {
-        let topo = Topology::new().expect("Failed to build host topology");
-
-        self.setup_topology_node(topo.span.as_raw_slice())?;
-
-        for (_, node) in topo.nodes {
-            self.setup_topology_node(node.span.as_raw_slice())?;
-        }
-
-        for (_, llc) in topo.all_llcs {
-            self.setup_topology_node(
-                Arc::<Llc>::into_inner(llc)
-                    .expect("missing llc")
-                    .span
-                    .as_raw_slice(),
-            )?;
-        }
-
-        for (_, core) in topo.all_cores {
-            self.setup_topology_node(
-                Arc::<Core>::into_inner(core)
-                    .expect("missing core")
-                    .span
-                    .as_raw_slice(),
-            )?;
-        }
-        for (_, cpu) in topo.all_cpus {
-            let mut mask = [0; 9];
-            mask[cpu.id.checked_shr(64).unwrap_or(0)] |= 1 << (cpu.id % 64);
-            self.setup_topology_node(&mask)?;
-        }
-
-        Ok(())
-    }
-
     fn run(&mut self, shutdown: Arc<AtomicBool>) -> Result<UserExitInfo> {
         let (res_ch, req_ch) = self.stats_server.channels();
 
@@ -335,9 +227,6 @@ impl<'a> Scheduler<'a> {
     }
 
     fn start(&mut self) -> Result<()> {
-        self.setup_arenas()?;
-        self.setup_topology()?;
-
         self.struct_ops = Some(scx_ops_attach!(self.skel, p2dq)?);
 
         if self.verbose > 1 {
@@ -429,6 +318,10 @@ fn main() -> Result<()> {
     let mut open_object = MaybeUninit::uninit();
     loop {
         let mut sched = Scheduler::init(&opts.sched, &opts.libbpf, &mut open_object, opts.verbose)?;
+        let task_size = std::mem::size_of::<types::task_p2dq>();
+        let arenalib = ArenaLib::init(sched.skel.object_mut(), task_size, *NR_CPU_IDS)?;
+        arenalib.setup()?;
+
         sched.start()?;
 
         if !sched.run(shutdown.clone())?.should_restart() {
